""" test knudsen number utility

    the knudsen number goes like:
        0   for larger  particles
        inf for smaller particles
"""

import pytest
import numpy as np
from particula.particles.properties import calculate_knudsen_number


def test_basic_calculation():
    """Test the basic calculation of the Knudsen number"""
    kn = calculate_knudsen_number(0.1, 0.05)
    assert kn == 2.0


def test_numpy_array_input():
    """Test when numpy arrays are provided for radius"""
    mfp = np.array([0.1])
    radius = np.array([0.05, 0.1])
    expected_results = np.array([2.0, 1.0])
    kn = calculate_knudsen_number(mfp, radius)
    np.testing.assert_array_equal(kn, expected_results)


def test_array_in_both_inputs():
    """Test when numpy arrays are provided for both radius and mfp"""
    mfp = np.array([0.1, 0.2])
    radius = np.array([0.05, 0.1, 0.8, 3])
    expected_results = mfp[np.newaxis, :] / radius[:, np.newaxis]
    kn = calculate_knudsen_number(mfp, radius)
    np.testing.assert_array_equal(kn, expected_results)


def test_zero_particle_radius():
    """Test when the particle radius is zero,
    which should raise an exception"""
    with pytest.raises(ZeroDivisionError):
        calculate_knudsen_number(0.1, 0.0)


def test_negative_inputs():
    """Test when negative inputs are provided to the function"""
    kn = calculate_knudsen_number(-0.1, -0.05)
    assert kn == 2.0


def test_invalid_type_inputs():
    """Test when invalid input types are provided to the function"""
    with pytest.raises(TypeError):
        calculate_knudsen_number("0.1", 0.05)  # Invalid string input

    with pytest.raises(TypeError):
        calculate_knudsen_number(0.1, "0.05")  # Invalid string input



""" testing getting constants from the get_constants.py file
"""

from particula.util.constants import (AVOGADRO_NUMBER,
                                      BOLTZMANN_CONSTANT,
                                      GAS_CONSTANT)


def test_constants():
    """ simple tests are conducted as follows:

            * see if GAS_CONSTANT maniuplation is good
    """

    assert (
        GAS_CONSTANT
        ==
        BOLTZMANN_CONSTANT * AVOGADRO_NUMBER
    )



"""Particle diffusion coefficient calculation."""

from typing import Union
import numpy as np
from numpy.typing import NDArray

from particula.util.constants import BOLTZMANN_CONSTANT
from particula.gas.properties.dynamic_viscosity import get_dynamic_viscosity
from particula.gas.properties.mean_free_path import molecule_mean_free_path
from particula.particles.properties.aerodynamic_mobility_module import (
    particle_aerodynamic_mobility,
)
from particula.particles.properties.slip_correction_module import (
    cunningham_slip_correction,
)
from particula.particles.properties.knudsen_number_module import (
    calculate_knudsen_number,
)


def particle_diffusion_coefficient(
    temperature: Union[float, NDArray[np.float64]],
    aerodynamic_mobility: Union[float, NDArray[np.float64]],
    boltzmann_constant: float = BOLTZMANN_CONSTANT,
) -> Union[float, NDArray[np.float64]]:
    """
    Calculate the diffusion coefficient of a particle.

    Arguments:
        temperature: The temperature at which the particle is
            diffusing, in Kelvin. Defaults to 298.15 K.
        boltzmann_constant: The Boltzmann constant. Defaults to the
            standard value of 1.380649 x 10^-23 J/K.
        aerodynamic_mobility: The aerodynamic mobility of
            the particle [m^2/s].

    Returns:
        The diffusion coefficient of the particle [m^2/s].
    """
    return boltzmann_constant * temperature * aerodynamic_mobility


def particle_diffusion_coefficient_via_system_state(
    particle_radius: Union[float, NDArray[np.float64]],
    temperature: float,
    pressure: float,
) -> Union[float, NDArray[np.float64]]:
    """
    Calculate the diffusion coefficient of a particle.

    Arguments:
        temperature: The temperature of the system in Kelvin (K).
        particle_radius: The radius of the particle in meters (m).
        pressure: The pressure of the system in Pascals (Pa).

    Returns:
        The diffusion coefficient of the particle in square meters per
        second (m²/s).
    """

    # Step 1: Calculate gas properties
    _dynamic_viscosity = get_dynamic_viscosity(temperature=temperature)
    _mean_free_path = molecule_mean_free_path(
        temperature=temperature,
        pressure=pressure,
        dynamic_viscosity=_dynamic_viscosity,
    )

    # Step 2: Particle properties in fluid
    _knudsen_number = calculate_knudsen_number(
        mean_free_path=_mean_free_path, particle_radius=particle_radius
    )
    _slip_correction_factor = cunningham_slip_correction(
        knudsen_number=_knudsen_number,
    )
    _aerodynamic_mobility = particle_aerodynamic_mobility(
        radius=particle_radius,
        slip_correction_factor=_slip_correction_factor,
        dynamic_viscosity=_dynamic_viscosity,
    )

    # Step 3: Calculate the particle diffusion coefficient
    return particle_diffusion_coefficient(
        temperature=temperature,
        aerodynamic_mobility=_aerodynamic_mobility,
    )



"""
Test for water activity calculation
"""

import numpy as np
from particula.activity.water_activity import (
    fixed_water_activity,
    biphasic_water_activity_point,
)


def test_fixed_water_activity():
    """test for fixed_water_activity function."""
    molar_mass_ratio = 18.016 / 250
    water_activity_desired = np.linspace(0.1, 0.9999, 1000)
    oxygen2carbon = 0.3
    density = 1500

    result = fixed_water_activity(
        water_activity=water_activity_desired,
        molar_mass_ratio=molar_mass_ratio,
        oxygen2carbon=oxygen2carbon,
        density=density,
    )

    assert isinstance(result, tuple)
    assert len(result) == 3


def test_biphasic_water_activity_point():
    """test for biphasic_water_activity_point function."""
    oxygen2carbon = 0.3
    hydrogen2carbon = 0.3
    molar_mass_ratio = 18 / 250

    # single entry
    activity_point = biphasic_water_activity_point(
        oxygen2carbon=oxygen2carbon,
        hydrogen2carbon=hydrogen2carbon,
        molar_mass_ratio=molar_mass_ratio,
        functional_group=None,
    )
    assert np.all(activity_point >= 0)

    # array entry
    oxygen2carbon_array = np.linspace(0, 0.6, 10)
    hydrogen2carbon_array = np.linspace(0, 0.6, 10)
    molar_mass_ratio_array = 18 / np.linspace(200, 500, 10)
    activity_point = biphasic_water_activity_point(
        oxygen2carbon=oxygen2carbon_array,
        hydrogen2carbon=hydrogen2carbon_array,
        molar_mass_ratio=molar_mass_ratio_array,
        functional_group=None,
    )
    assert np.all(activity_point >= 0)

    # edge cases
    activity_point = biphasic_water_activity_point(
        oxygen2carbon=1.0,
        hydrogen2carbon=1.0,
        molar_mass_ratio=1.0,
        functional_group=None,
    )
    assert np.all(activity_point >= 0)



"""Test for activity factories. Is the factory function working as expected,
creating the right activity strategy based on the input parameters,
and raising the right exceptions when needed.

The Builder is tested independently."""

import pytest
import numpy as np
from particula.particles.activity_factories import (
    ActivityFactory,
)
from particula.particles.activity_strategies import (
    ActivityIdealMass,
    ActivityIdealMolar,
    ActivityKappaParameter,
)


def test_mass_ideal_strategy_no_parameters():
    """Test factory function for mass_ideal strategy without parameters."""
    strategy = ActivityFactory().get_strategy("mass_ideal")
    assert isinstance(strategy, ActivityIdealMass)


def test_molar_ideal_strategy_with_parameters():
    """Test factory function for molar_ideal strategy with parameters."""
    parameters = {"molar_mass": np.array([100.0, 200.0, 300.0])}
    strategy = ActivityFactory().get_strategy("molar_ideal", parameters)
    assert isinstance(strategy, ActivityIdealMolar)
    np.testing.assert_allclose(
        strategy.molar_mass, parameters["molar_mass"], atol=1e-4
    )


def test_kappa_parameter_strategy_with_parameters():
    """
    Test factory function for kappa_parameter strategy with full parameters.
    """
    parameters = {
        "kappa": np.array([0.1, 0.2, 0.3]),
        "density": np.array([1000.0, 2000.0, 3000.0]),
        "density_units": "kg/m^3",
        "molar_mass": np.array([1.0, 2.0, 3.0]),
        "water_index": 0,
    }
    strategy = ActivityFactory().get_strategy("kappa_parameter", parameters)
    assert isinstance(strategy, ActivityKappaParameter)
    np.testing.assert_allclose(strategy.kappa, parameters["kappa"], atol=1e-4)
    np.testing.assert_allclose(
        strategy.density, parameters["density"], atol=1e-4
    )
    np.testing.assert_allclose(
        strategy.molar_mass, parameters["molar_mass"], atol=1e-4
    )
    assert strategy.water_index == parameters["water_index"]


def test_invalid_strategy_type():
    """Test factory function with an invalid strategy type."""
    with pytest.raises(ValueError) as excinfo:
        ActivityFactory().get_strategy("invalid_type")
    assert "Unknown strategy type: invalid_type" in str(excinfo.value)



"""
This module contains functions to calculate the phase separation of organic
compounds in water. The functions are based on the BAT model.

Gorkowski, K., Preston, T. C., Zuend, A. (2019).
Relative-humidity-dependent organic aerosol thermodynamics
Via an efficient reduced-complexity model.
Atmospheric Chemistry and Physics
https://doi.org/10.5194/acp-19-13383-2019
"""

from typing import Union
import numpy as np
from numpy.typing import NDArray

from particula.util.machine_limit import safe_exp, safe_log

MIN_SPREAD_IN_AW = 10**-6
Q_ALPHA_AT_1PHASE_AW = 0.99


def organic_water_single_phase(
    molar_mass_ratio: Union[int, float, list, np.ndarray]
) -> np.ndarray:
    """
    Convert the given molar mass ratio (MW water / MW organic) to a
    and oxygen2carbon value were above is a single phase with water and below
    phase separation is possible.

    Args:
        - molar_mass_ratio: The molar mass ratio with respect to water.

    Returns:
        - The single phase cross point.

    References:
        - Gorkowski, K., Preston, T. C., &#38; Zuend, A. (2019).
          Relative-humidity-dependent organic aerosol thermodynamics
          Via an efficient reduced-complexity model.
          Atmospheric Chemistry and Physics
          https://doi.org/10.5194/acp-19-13383-2019
    """
    # check inputs
    molar_mass_ratio = np.asarray(molar_mass_ratio, dtype=np.float64)

    return (
        0.205 / (1 + safe_exp(26.6 * (molar_mass_ratio - 0.12))) ** 0.843
        + 0.225
    )


# pylint: disable=too-many-locals
def find_phase_sep_index(activity_data: NDArray[np.float64]) -> dict:
    """
    This function finds phase separation using activity>1 and
    inflections in the activity curve data.
    In physical systems activity can not be above one and
    curve should be monotonic. Or else there will be phase separation.

    Args:
    - activity_data: A array of activity data.

    Returns:
    dict: A dictionary containing the following keys:
        - 'phase_sep_activity': Phase separation via activity
            (1 if there is phase separation, 0 otherwise)
        - 'phase_sep_curve': Phase separation via activity curvature
            (1 if there is phase separation, 0 otherwise)
        - 'index_phase_sep_starts': Index where phase separation starts
        - 'index_phase_sep_end': Index where phase separation ends
    """
    # check inputs
    activity_data = np.asarray(activity_data, dtype=np.float64)

    # Compute difference between consecutive elements in the array
    activity_diff = np.diff(activity_data)
    data_length = len(activity_data)

    # Check if the data length is more than 3
    if data_length > 3:
        min_value = np.min(activity_diff)
        max_value = np.max(activity_diff)

        # Check if the min and max differences have the same sign
        if np.sign(min_value) == np.sign(max_value):
            # If so, no phase separation via activity curvature
            phase_sep_curve = 0
            # find the index where the activity is closest to 1
            index_phase_sep_starts = np.argmin(np.abs(activity_data - 1))
            index_phase_sep_end = index_phase_sep_starts
        else:
            # If signs differ, phase separation via activity curvature occurs
            phase_sep_curve = 1
            # Find indices where the sign of the second derivative changes
            sign_changes = np.diff(np.sign(activity_diff))

            # The first index where a sign change occurs
            inflection_index = np.where(sign_changes)[0]  # all indices
            index_start = (
                inflection_index[0]
                if len(inflection_index) > 0
                else data_length
            )
            # The last index where a sign change occurs
            back_index = (
                inflection_index[-1]
                if len(inflection_index) > 0
                else data_length
            )

            # Check if first section of activity data is greater than 1
            if np.any(activity_data[:index_start] > 1):
                index_phase_sep_starts = np.argmin(
                    np.abs(activity_data[:index_start] - 1)
                )
            else:
                index_phase_sep_starts = index_start

            # Check if second section of activity data is greater than 1
            if np.any(activity_data[back_index:] > 1):
                index_phase_sep_end = (
                    np.argmin(np.abs(activity_data[back_index:] - 1))
                    + back_index
                )
            else:
                index_phase_sep_end = back_index
    else:
        phase_sep_activity = activity_data
        phase_sep_curve = 0
        index_phase_sep_starts = np.nan
        index_phase_sep_end = np.nan

    # Assign phase separation via activity based on data being greater than 1
    phase_sep_activity = 1 if sum(activity_data > 1) else 0
    return {
        "phase_sep_activity": phase_sep_activity,
        "phase_sep_curve": phase_sep_curve,
        "index_phase_sep_starts": index_phase_sep_starts,
        "index_phase_sep_end": index_phase_sep_end,
    }


def find_phase_separation(
    activity_water: NDArray[np.float64], activity_org: NDArray[np.float64]
) -> dict:
    """
    This function checks for phase separation in each activity curve.

    Args:
    - activity_water (np.array): A numpy array of water activity values.
    - activity_org (np.array): A numpy array of organic activity values.

    Returns:
    dict: A dictionary containing the following keys:
        - 'phase_sep_check': An integer indicating whether phase separation
                is present (1) or not (0).
        - 'lower_seperation_index': The index of the lower separation point
                in the activity curve.
        - 'upper_seperation_index': The index of the upper separation point in
                the activity curve.
        - 'matching_upper_seperation_index': The index where the difference
                between activity_water_beta and match_a_w is greater than 0.
        - 'lower_seperation': The value of water activity at the lower
                separation point.
        - 'upper_seperation': The value of water activity at the upper
                separation point.
        - 'matching_upper_seperation': The value of water activity at the
                matching upper separation point.
    """

    # check for phase separation in each activity curve
    water_sep = find_phase_sep_index(activity_water)
    organic_sep = find_phase_sep_index(activity_org)

    # gather all the indexes into a list for easier access
    indexes = [
        water_sep["index_phase_sep_starts"],
        water_sep["index_phase_sep_end"],
        organic_sep["index_phase_sep_starts"],
        organic_sep["index_phase_sep_end"],
    ]

    # If there is a phase separation curve in the water activity data
    if water_sep["phase_sep_curve"] == 1:
        phase_sep_check = 1

        # Check for the direction of the curve (increasing or decreasing)
        if activity_water[0] < activity_water[-1]:  # increasing a_w with index
            # find the min and max indexes
            lower_seperation_index = min(indexes)
            upper_seperation_index = max(indexes)
            match_a_w = activity_water[upper_seperation_index]

            # start from the lower_seperation_index and find the index where
            # the difference between activity_water and match_a_w changes sign
            match_slice = np.sign(
                match_a_w - activity_water[lower_seperation_index:]
            )
            match_index_prime = np.where(match_slice == -1)
            if len(match_index_prime[0]) == 0:
                match_index_prime = lower_seperation_index
            else:
                match_index_prime = (
                    match_index_prime[0][0] + lower_seperation_index
                )
        else:  # decreasing a_w with index
            # find the min and max indexes
            lower_seperation_index = max(indexes)
            upper_seperation_index = min(indexes)
            match_a_w = activity_water[upper_seperation_index]

            # start from the lower_seperation_index and find the index where
            # the difference between activity_water and match_a_w changes sign
            match_slice = np.sign(
                activity_water[:lower_seperation_index] - match_a_w
            )
            match_index_prime = np.where(match_slice == -1)
            if len(match_index_prime[0]) == 0:
                match_index_prime = lower_seperation_index
            else:
                match_index_prime = match_index_prime[0][0]

    else:
        upper_seperation_index = 2
        lower_seperation_index = 2
        match_index_prime = 2
        phase_sep_check = 0  # no phase seperation

    return {
        "phase_sep_check": phase_sep_check,
        "lower_seperation_index": lower_seperation_index,
        "upper_seperation_index": upper_seperation_index,
        "matching_upper_seperation_index": match_index_prime,
        "lower_seperation": activity_water[lower_seperation_index],
        "upper_seperation": activity_water[upper_seperation_index],
        "matching_upper_seperation": activity_water[match_index_prime],
    }


def q_alpha(
    seperation_activity: NDArray[np.float64],
    activities: NDArray[np.float64],
) -> np.ndarray:
    """
    This function calculates the q_alpha value using a squeezed logistic
        function.

    Args:
        - seperation_activity (np.array): A numpy array of values representing
            the separation activity.
        - activities (np.array): A numpy array of activity values.

    Returns:
        np.array: The q_alpha value.

    Notes:
        - The q_alpha value represents the transfer from
            q_alpha ~0 to q_alpha ~1.
        - The function uses a sigmoid curve parameter to calculate the
            q_alpha value.
    """
    # check inputs
    if seperation_activity == 0:
        return np.ones_like(activities)

    # spread in transfer from 50/50 point
    delta_seperation = 1 - seperation_activity

    # check min value allowed
    above_min_delta_seperation_value = delta_seperation > MIN_SPREAD_IN_AW
    delta_seperation = (
        delta_seperation * above_min_delta_seperation_value
        + (not above_min_delta_seperation_value) * MIN_SPREAD_IN_AW
    )

    # calculate curve parameter of sigmoid
    sigmoid_curve_parameter = (
        safe_log(1 / (1 - Q_ALPHA_AT_1PHASE_AW) - 1) / delta_seperation
    )

    # calculate q_alpha return value
    return 1 - 1 / (
        1
        + safe_exp(
            sigmoid_curve_parameter
            * (activities - seperation_activity + delta_seperation)
        )
    )



""" testing the utility calculating reduced quantities
"""

import numpy as np
import pytest
from particula.util.reduced_quantity import (
    reduced_value, reduced_self_broadcast)


def test_reduced_value_scalar():
    """ Test that the reduced value is calculated correctly."""
    assert reduced_value(4, 2) == 4 * 2 / (4 + 2), "Failed for scalar inputs"


def test_reduced_value_array():
    """ Test that the reduced value is calculated correctly."""
    alpha = np.array([2, 4, 6])
    beta = np.array([3, 6, 9])
    expected = alpha * beta / (alpha + beta)
    result = reduced_value(alpha, beta)
    assert np.allclose(result, expected), "Failed for array inputs"


def test_reduced_value_zero_division():
    """ Test division by zero handling."""
    alpha = np.array([0, 2, 0])
    beta = np.array([0, 0, 2])
    # Expect zeros where division by zero occurs
    expected = np.array([0, 0, 0])
    result = reduced_value(alpha, beta)
    assert np.array_equal(result, expected), "Failed handling division by zero"


def test_reduced_value_shape_mismatch():
    """ Test error handling for shape mismatch."""
    alpha = np.array([1, 2])
    beta = np.array([1, 2, 3])
    with pytest.raises(ValueError):
        reduced_value(alpha, beta)


def test_reduced_value_negative_values():
    """ Test that the reduced value is calculated correctly."""
    alpha = np.array([-1, -2])
    beta = np.array([-3, -4])
    expected = alpha * beta / (alpha + beta)
    result = reduced_value(alpha, beta)
    assert np.allclose(result, expected), "Failed for negative values"


def test_reduced_value_one_element():
    """ Test with one element in array"""
    alpha = np.array([5])
    beta = np.array([10])
    expected = alpha * beta / (alpha + beta)
    result = reduced_value(alpha, beta)
    assert np.allclose(result, expected), "Failed for single element arrays"


def test_reduced_self_broadcast_typical():
    """ Test that the reduced self broadcast is calculated correctly."""
    alpha_array = np.array([1, 2, 3])
    expected_result = np.array([
        [1 * 1 / (1 + 1), 1 * 2 / (1 + 2), 1 * 3 / (1 + 3)],
        [2 * 1 / (2 + 1), 2 * 2 / (2 + 2), 2 * 3 / (2 + 3)],
        [3 * 1 / (3 + 1), 3 * 2 / (3 + 2), 3 * 3 / (3 + 3)]
    ])
    result = reduced_self_broadcast(alpha_array)
    assert np.allclose(
        result, expected_result), "Test failed for typical input"


def test_reduced_self_broadcast_empty():
    """ Test with empty input"""
    alpha_array = np.array([])
    result = reduced_self_broadcast(alpha_array)
    assert result.shape == (0, 0), "Test failed for empty input"


def test_reduced_self_broadcast_zero_elements():
    """ Test with zero elements"""
    alpha_array = np.array([0, 0, 0])
    expected_result = np.array([
        [0, 0, 0],
        [0, 0, 0],
        [0, 0, 0]
    ])
    result = reduced_self_broadcast(alpha_array)
    assert np.allclose(
        result, expected_result), "Test failed for zero elements input"


def test_reduced_self_broadcast_one_element():
    """ Test with a single element array"""
    alpha_array = np.array([4])
    expected_result = np.array([
        [4 * 4 / (4 + 4)]
    ])
    result = reduced_self_broadcast(alpha_array)
    assert np.allclose(
        result, expected_result), "Test failed for one element input"


def test_reduced_self_broadcast_negative_elements():
    """ Test with negative elements"""
    alpha_array = np.array([-1, -2, -3])
    expected_result = np.array([
        [-1 * -1 / (-1 - 1), -1 * -2 / (-1 - 2), -1 * -3 / (-1 - 3)],
        [-2 * -1 / (-2 - 1), -2 * -2 / (-2 - 2), -2 * -3 / (-2 - 3)],
        [-3 * -1 / (-3 - 1), -3 * -2 / (-3 - 2), -3 * -3 / (-3 - 3)]
    ])
    result = reduced_self_broadcast(alpha_array)
    assert np.allclose(
        result, expected_result), "Test failed for negative elements input"



"""
Test for activity_coefficients function.
"""

import numpy as np
from particula.activity.activity_coefficients import bat_activity_coefficients


def test_activity_coefficients():
    """test for activity_coefficents function."""
    molar_mass_ratio = 18.016 / 250
    organic_mole_fraction = np.linspace(0.1, 1, 10)
    oxygen2carbon = 0.3
    density = 2500

    activity_coefficients = bat_activity_coefficients(
        molar_mass_ratio=molar_mass_ratio,
        organic_mole_fraction=organic_mole_fraction,
        oxygen2carbon=oxygen2carbon,
        density=density,
        functional_group=None,
    )
    assert np.all(activity_coefficients[0] >= 0)
    assert np.all(activity_coefficients[1] >= 0)
    assert np.all(activity_coefficients[2] >= 0)
    assert np.all(activity_coefficients[3] >= 0)


def test_activity_coefficients_with_functional_group():
    """test for activity_coefficents function with functional group."""
    molar_mass_ratio = 18.016 / 250
    organic_mole_fraction = np.linspace(0.1, 1, 10)
    oxygen2carbon = 0.3
    density = 2500
    functional_group = "alcohol"

    activity_coefficients = bat_activity_coefficients(
        molar_mass_ratio=molar_mass_ratio,
        organic_mole_fraction=organic_mole_fraction,
        oxygen2carbon=oxygen2carbon,
        density=density,
        functional_group=functional_group,
    )
    assert np.all(activity_coefficients[0] >= 0)
    assert np.all(activity_coefficients[1] >= 0)
    assert np.all(activity_coefficients[2] >= 0)
    assert np.all(activity_coefficients[3] >= 0)


def test_activity_coefficients_edge_cases():
    """test for activity_coefficents function with edge cases."""
    molar_mass_ratio = 18.016 / 250
    organic_mole_fraction = np.array([1e-12, 1.0])
    oxygen2carbon = 0.3
    density = 2500

    activity_coefficients = bat_activity_coefficients(
        molar_mass_ratio=molar_mass_ratio,
        organic_mole_fraction=organic_mole_fraction,
        oxygen2carbon=oxygen2carbon,
        density=density,
        functional_group=None,
    )
    assert np.all(activity_coefficients[0] >= 0)
    assert np.all(activity_coefficients[1] >= 0)
    assert np.all(activity_coefficients[2] >= 0)
    assert np.all(activity_coefficients[3] >= 0)


def test_activity_coefficients_with_arrays():
    """test for activity_coefficents function with np.array inputs."""
    molar_mass_ratio = np.array([18.016 / 250, 18.016 / 200])
    organic_mole_fraction = np.array([0.2, 0.8])
    oxygen2carbon = np.array([0.3, 0.4])
    density = np.array([2500, 2400])

    activity_coefficients = bat_activity_coefficients(
        molar_mass_ratio=molar_mass_ratio,
        organic_mole_fraction=organic_mole_fraction,
        oxygen2carbon=oxygen2carbon,
        density=density,
        functional_group=None,
    )
    assert np.all(activity_coefficients[0] >= 0)
    assert np.all(activity_coefficients[1] >= 0)
    assert np.all(activity_coefficients[2] >= 0)
    assert np.all(activity_coefficients[3] >= 0)


def test_activity_coefficients_with_mixed_inputs():
    """test for activity_coefficents function with mixed inputs."""
    molar_mass_ratio = 18.016 / 250
    organic_mole_fraction = np.array([0.1, 0.9])
    oxygen2carbon = 0.3
    density = np.array([2500, 2400])

    activity_coefficients = bat_activity_coefficients(
        molar_mass_ratio=molar_mass_ratio,
        organic_mole_fraction=organic_mole_fraction,
        oxygen2carbon=oxygen2carbon,
        density=density,
        functional_group=None,
    )
    assert np.all(activity_coefficients[0] >= 0)
    assert np.all(activity_coefficients[1] >= 0)
    assert np.all(activity_coefficients[2] >= 0)
    assert np.all(activity_coefficients[3] >= 0)



"""
Calculate the geometric collision kernel Γ₁₂ (or K₁₂) based on turbulent
DNS simulations.

- Ayala, O., Rosa, B., & Wang, L. P. (2008). Effects of turbulence on
    the geometric collision rate of sedimenting droplets. Part 2.
    Theory and parameterization. New Journal of Physics, 10.
    https://doi.org/10.1088/1367-2630/10/7/075016
"""

from typing import Union
import numpy as np
from numpy.typing import NDArray

from particula.util.validate_inputs import validate_inputs
from particula.dynamics.coagulation.turbulent_dns_kernel.radial_velocity_module import (
    get_radial_relative_velocity_ao2008,
)
from particula.dynamics.coagulation.turbulent_dns_kernel.g12_radial_distribution_ao2008 import (
    get_g12_radial_distribution_ao2008,
)
from particula.dynamics.coagulation.turbulent_dns_kernel.sigma_relative_velocity_ao2008 import (
    get_relative_velocity_variance,
)
from particula.particles.properties import (
    get_particle_inertia_time,
    get_particle_settling_velocity_via_inertia,
    get_particle_reynolds_number,
    get_stokes_number,
    cunningham_slip_correction,
    calculate_knudsen_number,
)
from particula.gas.properties import (
    get_dynamic_viscosity,
    molecule_mean_free_path,
    get_kinematic_viscosity,
    get_eulerian_integral_length,
    get_lagrangian_integral_time,
    get_taylor_microscale,
    get_fluid_rms_velocity,
    get_normalized_accel_variance_ao2008,
    get_kolmogorov_length,
    get_kolmogorov_velocity,
    get_kolmogorov_time,
)
from particula.util.constants import STANDARD_GRAVITY


@validate_inputs(
    {
        "particle_radius": "positive",
        "velocity_dispersion": "positive",
        "particle_inertia_time": "positive",
    }
)
def get_kernel_ao2008(
    particle_radius: Union[float, NDArray[np.float64]],
    velocity_dispersion: Union[float, NDArray[np.float64]],
    particle_inertia_time: Union[float, NDArray[np.float64]],
    stokes_number: Union[float, NDArray[np.float64]],
    kolmogorov_length_scale: float,
    reynolds_lambda: float,
    normalized_accel_variance: float,
    kolmogorov_velocity: float,
    kolmogorov_time: float,
) -> Union[float, NDArray[np.float64]]:
    # pylint: disable=too-many-arguments, too-many-positional-arguments
    """
    Get the geometric collision kernel Γ₁₂.

        Γ₁₂ = 2π R² ⟨ |w_r| ⟩ g₁₂

    - R = a₁ + a₂ (collision radius)
    - ⟨ |w_r| ⟩ : Radial relative velocity, computed using
        `get_radial_relative_velocity_ao2008`
    - g₁₂ : Radial distribution function, computed using
        `g12_radial_distribution`
    - radius << η (Kolmogorov length scale)
    - ρ_w >> ρ (water density much greater than air density)
    - Sv > 1 (Stokes number sufficiently large)

    Arguments:
    ----------
        - particle_radius : Particle radius [m].
        - velocity_dispersion : Velocity dispersion [m/s].
        - particle_inertia_time : Particle inertia time [s].
        - stokes_number : Stokes number [-].
        - kolmogorov_length_scale : Kolmogorov length scale [m].
        - reynolds_lambda : Reynolds number [-].
        - normalized_accel_variance : Normalized acceleration variance [-].
        - kolmogorov_velocity : Kolmogorov velocity [m/s].
        - kolmogorov_time : Kolmogorov time [s].

    Returns:
    --------
        - Collision kernel Γ₁₂ [m³/s].

    References:
    -----------
    - Ayala, O., Rosa, B., & Wang, L. P. (2008). Effects of turbulence on
        the geometric collision rate of sedimenting droplets. Part 2.
        Theory and parameterization. New Journal of Physics, 10.
        https://doi.org/10.1088/1367-2630/10/7/075016
    """
    collision_radius = (
        particle_radius[:, np.newaxis] + particle_radius[np.newaxis, :]
    )

    # Compute radial relative velocity ⟨ |w_r| ⟩
    wr = get_radial_relative_velocity_ao2008(
        velocity_dispersion, particle_inertia_time
    )

    # Compute radial distribution function g₁₂
    g12 = get_g12_radial_distribution_ao2008(
        particle_radius,
        stokes_number,
        kolmogorov_length_scale,
        reynolds_lambda,
        normalized_accel_variance,
        kolmogorov_velocity,
        kolmogorov_time,
    )

    # Compute collision kernel Γ₁₂
    gamma_12 = 2 * np.pi * collision_radius**2 * wr * g12

    return gamma_12


def get_kernel_ao2008_via_system_state(
    particle_radius: Union[float, NDArray[np.float64]],
    particle_density: Union[float, NDArray[np.float64]],
    fluid_density: float,
    temperature: float,
    re_lambda: float,
    relative_velocity: Union[float, NDArray[np.float64]],
    turbulent_dissipation: float,
) -> Union[float, NDArray[np.float64]]:
    # pylint: disable=too-many-arguments, too-many-positional-arguments
    # pylint: disable=too-many-locals
    """
    Compute the geometric collision kernel (Γ₁₂) under the AO2008 formulation.
    This function orchestrates the calculation of the geometric collision
    kernel by deriving necessary fluid, turbulence, and particle parameters
    from the provided system state. The returned value (or array) represents
    the collision kernel, Γ₁₂ [m³/s], which describes collision frequency
    under turbulence.
    Arguments:
    ----------
    - particle_radius : Radius of the particles [m]. If an array is given, it
        is assumed to represent multiple particle sizes.
    - particle_density : Density of the particles [kg/m³]. Must match the
        dimensionality of `particle_radius` if both are arrays.
    - fluid_density : Density of the surrounding fluid [kg/m³].
    - temperature : Temperature of the fluid [K].
    - re_lambda : Turbulent Reynolds number based on the Taylor microscale.
    - relative_velocity : Mean relative velocity between the particle and fluid
        [m/s]. Can be a single value or an array of the same dimensionality as
        `particle_radius`.
    - turbulent_dissipation : Turbulent kinetic energy dissipation rate
        [m²/s³].
    Returns:
    --------
    - The geometric collision kernel Γ₁₂ [m³/s]. If inputs are scalars, returns
        a float. If inputs are arrays, returns a numpy array of collision
        kernels.
    Notes:
    -----
    This function does the following:
    1. Calculates fluid properties (dynamic, kinematic viscosity,
        mean free path).
    2. Determines slip correction factors (Knudsen number, Cunningham factor).
    3. Computes particle inertia times and settling velocities.
    4. Estimates relevant turbulence scales (Kolmogorov, Taylor,
        integral scales).
    5. Calculates velocity variance and auxiliary terms, e.g. Stokes number.
    6. Calls `get_kernel_ao2008` with all the necessary inputs to get the final
       collision kernel.
    References:
    -----------
    - Ayala, O., Rosa, B., & Wang, L. P. (2008). Effects of turbulence on
        the geometric collision rate of sedimenting droplets. Part 2.
        Theory and parameterization. New Journal of Physics, 10.
        https://doi.org/10.1088/1367-2630/10/7/075016
    """
    # 1. Basic fluid properties
    dynamic_viscosity = get_dynamic_viscosity(temperature)
    kinematic_viscosity = get_kinematic_viscosity(
        dynamic_viscosity=dynamic_viscosity, fluid_density=fluid_density
    )
    mean_free_path = molecule_mean_free_path(
        temperature=temperature, dynamic_viscosity=dynamic_viscosity
    )

    # 2. Slip correction factors
    knudsen_number = calculate_knudsen_number(
        mean_free_path=mean_free_path, particle_radius=particle_radius
    )
    slip_correction_factor = cunningham_slip_correction(knudsen_number)

    # Handle radius addition properly for arrays
    collisional_radius = (
        particle_radius[:, np.newaxis] + particle_radius[np.newaxis, :]
        if isinstance(particle_radius, np.ndarray)
        else 2.0 * particle_radius
    )

    # 3. Particle inertia and settling velocity
    particle_inertia_time = get_particle_inertia_time(
        particle_radius=particle_radius,
        particle_density=particle_density,
        fluid_density=fluid_density,
        kinematic_viscosity=kinematic_viscosity,
        relative_velocity=relative_velocity,
    )
    particle_settling_velocity = get_particle_settling_velocity_via_inertia(
        particle_inertia_time=particle_inertia_time,
        gravitational_acceleration=STANDARD_GRAVITY,
        slip_correction_factor=slip_correction_factor,
    )

    # 4. Turbulence scales
    fluid_rms_velocity = get_fluid_rms_velocity(
        re_lambda=re_lambda,
        kinematic_viscosity=kinematic_viscosity,
        turbulent_dissipation=turbulent_dissipation,
    )
    taylor_microscale = get_taylor_microscale(
        fluid_rms_velocity=fluid_rms_velocity,
        kinematic_viscosity=kinematic_viscosity,
        turbulent_dissipation=turbulent_dissipation,
    )
    eulerian_integral_length = get_eulerian_integral_length(
        fluid_rms_velocity=fluid_rms_velocity,
        turbulent_dissipation=turbulent_dissipation,
    )
    lagrangian_integral_time = get_lagrangian_integral_time(
        fluid_rms_velocity=fluid_rms_velocity,
        turbulent_dissipation=turbulent_dissipation,
    )

    # 5. Relative velocity variance
    velocity_dispersion = get_relative_velocity_variance(
        fluid_rms_velocity=fluid_rms_velocity,
        collisional_radius=collisional_radius,
        particle_inertia_time=particle_inertia_time,
        particle_velocity=particle_settling_velocity,
        taylor_microscale=taylor_microscale,
        eulerian_integral_length=eulerian_integral_length,
        lagrangian_integral_time=lagrangian_integral_time,
    )

    # 6. Additional turbulence-based quantities
    kolmogorov_time = get_kolmogorov_time(
        kinematic_viscosity=kinematic_viscosity,
        turbulent_dissipation=turbulent_dissipation,
    )
    stokes_number = get_stokes_number(
        particle_inertia_time=particle_inertia_time,
        kolmogorov_time=kolmogorov_time,
    )
    kolmogorov_length_scale = get_kolmogorov_length(
        kinematic_viscosity=kinematic_viscosity,
        turbulent_dissipation=turbulent_dissipation,
    )
    reynolds_lambda = get_particle_reynolds_number(
        particle_radius=particle_radius,
        particle_velocity=particle_settling_velocity,
        kinematic_viscosity=kinematic_viscosity,
    )
    normalized_accel_variance = get_normalized_accel_variance_ao2008(
        re_lambda=reynolds_lambda,
    )
    kolmogorov_velocity = get_kolmogorov_velocity(
        kinematic_viscosity=kinematic_viscosity,
        turbulent_dissipation=turbulent_dissipation,
    )

    # 7. Final collision kernel
    return get_kernel_ao2008(
        particle_radius=particle_radius,
        velocity_dispersion=velocity_dispersion,
        particle_inertia_time=particle_inertia_time,
        stokes_number=stokes_number,
        kolmogorov_length_scale=kolmogorov_length_scale,
        reynolds_lambda=reynolds_lambda,
        normalized_accel_variance=normalized_accel_variance,
        kolmogorov_velocity=kolmogorov_velocity,
        kolmogorov_time=kolmogorov_time,
    )



"""
Test psi function from Ayala et al. (2008).
"""

import pytest
import numpy as np
from particula.dynamics.coagulation.turbulent_dns_kernel.psi_ao2008 import (
    get_psi_ao2008,
)


def test_get_psi_ao2008_scalar():
    """
    Test get_psi_ao2008 with scalar inputs.
    """
    alpha = 2.0  # Turbulence parameter [-]
    phi = 1.0  # Characteristic velocity [m/s]
    particle_inertia_time = 0.05  # [s]
    particle_velocity = 0.1  # [m/s]

    expected = 1 / (
        (1 / particle_inertia_time) + (1 / alpha) + (particle_velocity / phi)
    ) - (
        particle_velocity
        / (
            2
            * phi
            * (
                (1 / particle_inertia_time)
                + (1 / alpha)
                + (particle_velocity / phi)
            )
            ** 2
        )
    )

    result = get_psi_ao2008(
        alpha, phi, particle_inertia_time, particle_velocity
    )

    assert np.isclose(
        result, expected, atol=1e-10
    ), f"Expected {expected}, but got {result}"


def test_get_psi_ao2008_array():
    """
    Test get_psi_ao2008 with NumPy array inputs.
    """
    alpha = 2.0  # Turbulence parameter [-]
    phi = 1.0  # Characteristic velocity [m/s]
    particle_inertia_time = np.array([0.05, 0.1, 0.2])  # [s]
    particle_velocity = np.array([0.1, 0.2, 0.3])  # [m/s]

    denominator = (
        (1 / particle_inertia_time) + (1 / alpha) + (particle_velocity / phi)
    )
    expected = 1 / denominator - (
        particle_velocity / (2 * phi * denominator**2)
    )

    result = get_psi_ao2008(
        alpha, phi, particle_inertia_time, particle_velocity
    )

    assert (
        result.shape == expected.shape
    ), f"Expected shape {expected.shape}, but got {result.shape}"
    assert np.allclose(
        result, expected, atol=1e-10
    ), f"Expected {expected}, but got {result}"


def test_get_psi_ao2008_invalid_inputs():
    """
    Test that get_psi_ao2008 raises validation errors for invalid inputs.
    """
    alpha = 2.0  # [-]
    phi = 1.0  # [m/s]
    particle_inertia_time = np.array([0.05, 0.1, 0.2])  # [s]
    particle_velocity = np.array([0.1, 0.2, 0.3])  # [m/s]

    with pytest.raises(ValueError):
        get_psi_ao2008(
            -alpha, phi, particle_inertia_time, particle_velocity
        )  # Negative alpha

    with pytest.raises(ValueError):
        get_psi_ao2008(
            alpha, -phi, particle_inertia_time, particle_velocity
        )  # Negative phi

    with pytest.raises(ValueError):
        get_psi_ao2008(
            alpha, phi, -particle_inertia_time, particle_velocity
        )  # Negative inertia time

    with pytest.raises(ValueError):
        get_psi_ao2008(
            alpha, phi, particle_inertia_time, -particle_velocity
        )  # Negative velocity


def test_get_psi_ao2008_edge_cases():
    """
    Test get_psi_ao2008 with extreme values such as very small or very large
    values.
    """
    alpha = 2.0  # [-]
    phi = 1.0  # [m/s]
    particle_inertia_time = np.array(
        [1e-6, 1e-3, 10.0]
    )  # Very small and large inertia values
    particle_velocity = np.array(
        [1e-6, 1e-3, 10.0]
    )  # Very small and large velocities

    result = get_psi_ao2008(
        alpha, phi, particle_inertia_time, particle_velocity
    )

    assert np.all(result >= 0), "Expected all values to be non-negative"
    assert np.isfinite(result).all(), "Expected all values to be finite"



""" calculating the mean free path of air

    The mean free path is the average distance
    traveled by a molecule between collisions
    with other molecules present in a medium (air).

    The expeected mean free path of air is approx.
    65 nm at 298 K and 101325 Pa.

"""

import logging
from typing import Union, Optional
from numpy.typing import NDArray
import numpy as np
from particula.util.constants import (
    GAS_CONSTANT,
    MOLECULAR_WEIGHT_AIR,
)  # type: ignore
from particula.gas.properties.dynamic_viscosity import (
    get_dynamic_viscosity,
)

logger = logging.getLogger("particula")  # get instance of logger


def molecule_mean_free_path(
    molar_mass: Union[
        float, NDArray[np.float64]
    ] = MOLECULAR_WEIGHT_AIR,  # type: ignore
    temperature: float = 298.15,
    pressure: float = 101325,
    dynamic_viscosity: Optional[float] = None,
) -> Union[float, NDArray[np.float64]]:
    """
    Calculate the mean free path of a gas molecule in air based on the
    temperature, pressure, and molar mass of the gas. The mean free path
    is the average distance traveled by a molecule between collisions with
    other molecules present in a medium (air).

    Args:
    -----
    - molar_mass (Union[float, NDArray[np.float64]]): The molar mass
    of the gas molecule [kg/mol]. Default is the molecular weight of air.
    - temperature (float): The temperature of the gas [K]. Default is 298.15 K.
    - pressure (float): The pressure of the gas [Pa]. Default is 101325 Pa.
    - dynamic_viscosity (Optional[float]): The dynamic viscosity of the gas
    [Pa*s]. If not provided, it will be calculated based on the temperature.

    Returns:
    --------
    - Union[float, NDArray[np.float64]]: The mean free path of the gas molecule
    in meters (m).

    References:
    ----------
    - https://en.wikipedia.org/wiki/Mean_free_path
    """
    # check inputs are positive
    if temperature <= 0:
        logger.error("Temperature must be positive [Kelvin]")
        raise ValueError("Temperature must be positive [Kelvin]")
    if pressure <= 0:
        logger.error("Pressure must be positive [Pascal]")
        raise ValueError("Pressure must be positive [Pascal]")
    if np.any(molar_mass <= 0):
        logger.error("Molar mass must be positive [kg/mol]")
        raise ValueError("Molar mass must be positive [kg/mol]")
    if dynamic_viscosity is None:
        dynamic_viscosity = get_dynamic_viscosity(temperature)
    gas_constant = float(GAS_CONSTANT)

    return (2 * dynamic_viscosity / pressure) / (
        8 * molar_mass / (np.pi * gas_constant * temperature)
    ) ** 0.5



"""
Tests for the phi_function module.
"""

import pytest
import numpy as np

from particula.dynamics.coagulation.turbulent_dns_kernel.phi_ao2008 import (
    get_phi_ao2008
)


def test_get_phi_ao2008_scalar():
    """
    Test get_phi_ao2008 with scalar inputs.
    """
    alpha = 2.0  # Turbulence parameter [-]
    phi = 1.0  # Characteristic velocity [m/s]
    particle_inertia_time = np.array([0.05, 0.02])  # [s]
    particle_velocity = np.array([0.3, 0.1])  # [m/s]

    expected_shape = (2, 2)
    result = get_phi_ao2008(
        alpha, phi, particle_inertia_time, particle_velocity
    )

    assert (
        result.shape == expected_shape
    ), f"Expected shape {expected_shape}, but got {result.shape}"
    assert np.all(result >= 0), "Expected all values to be non-negative"


def test_get_phi_ao2008_array():
    """
    Test get_phi_ao2008 with NumPy array inputs.
    """
    alpha = 2.0  # Turbulence parameter [-]
    phi = 1.0  # Characteristic velocity [m/s]
    particle_inertia_time = np.array([0.05, 0.1, 0.2])  # [s]
    particle_velocity = np.array([0.1, 0.2, 0.3])  # [m/s]

    expected_shape = (3, 3)
    result = get_phi_ao2008(
        alpha, phi, particle_inertia_time, particle_velocity
    )

    assert (
        result.shape == expected_shape
    ), f"Expected shape {expected_shape}, but got {result.shape}"
    assert np.all(result >= 0), "Expected all values to be non-negative"


def test_get_phi_ao2008_invalid_inputs():
    """
    Test that get_phi_ao2008 raises validation errors for invalid inputs.
    """
    alpha = 2.0  # [-]
    phi = 1.0  # [m/s]
    particle_inertia_time = np.array([0.05, 0.1, 0.5])  # [s]
    particle_velocity = np.array([0.1, 0.2, 0.6])  # [m/s]

    with pytest.raises(ValueError):
        get_phi_ao2008(
            -alpha, phi, particle_inertia_time, particle_velocity
        )  # Negative alpha

    with pytest.raises(ValueError):
        get_phi_ao2008(
            alpha, -phi, particle_inertia_time, particle_velocity
        )  # Negative phi

    with pytest.raises(ValueError):
        get_phi_ao2008(
            alpha, phi, -particle_inertia_time, particle_velocity
        )  # Negative inertia time

    with pytest.raises(ValueError):
        get_phi_ao2008(
            alpha, phi, particle_inertia_time, -particle_velocity
        )  # Negative velocity


def test_get_phi_ao2008_edge_cases():
    """
    Test get_phi_ao2008 with extreme values such as very small or very large
    values.
    """
    alpha = 2.0  # [-]
    phi = 1.0  # [m/s]
    particle_inertia_time = np.array(
        [5e-6, 1e-3, 20.0]
    )  # Very small and large inertia values
    particle_velocity = np.array(
        [2e-6, 1e-3, 20.0]
    )  # Very small and large velocities

    result = get_phi_ao2008(
        alpha, phi, particle_inertia_time, particle_velocity
    )

    assert np.all(result >= 0), "Expected all values to be non-negative"
    assert np.isfinite(result).all(), "Expected all values to be finite"



"""
Pint Unit Conversion Wrapper

This is the only module that uses the pint library.
"""

from typing import Optional

try:
    import pint
    UNIT_REGISTRY = pint.UnitRegistry()
except ImportError:
    UNIT_REGISTRY = None


def convert_units(
    old: str,
    new: str,
    value: Optional[float] = None,
) -> float:
    """Generic wrapper for pint to convert units

    Args:
        - old : old units defined by pint, e.g., "m", "ft", "Km", "kg/m^3"
        - new : new units defined by pint, e.g., "m", "ft", "Km", "kg/m^3"
        - value : value to convert, needed for units with and offset
            e.g., "degC"

    Returns:
        - float : conversion multiplier from old to new units. If value is
            provided, it returns the converted value in the new units.

    Raises:
        ImportError: if pint is not installed

    Examples:
    ``` py title="Example Usage"
    conversion_multipliter = convert_units("ug/m^3", "kg/m^3")
    # 1e-9
    ```
    """
    if UNIT_REGISTRY is None:
        raise ImportError(
            "Install pint to use unit conversion features: pip install pint"
        )

    offset_units = ["degC", "degF", "degR", "degK"]
    if old in offset_units or value is not None:
        value = value if value is not None else 0
        old = UNIT_REGISTRY.Quantity(value, old)
    else:
        old = UNIT_REGISTRY.Quantity(old)  # multiplicative shift

    new = UNIT_REGISTRY.Quantity(new)
    result = old.to(new).magnitude  # get the new value without units
    return float(result)



"""Test the species factory module, does build the gas species object."""

import unittest
from particula.gas.species_factories import GasSpeciesFactory
from particula.gas.species_builders import (
    GasSpeciesBuilder,
    PresetGasSpeciesBuilder,
)
from particula.gas.species import GasSpecies
from particula.gas.vapor_pressure_strategies import (
    ConstantVaporPressureStrategy,
)


class TestGasSpeciesFactory(unittest.TestCase):
    """Test the GasSpeciesFactory class."""
    def setUp(self):
        """Set up the test case."""
        self.factory = GasSpeciesFactory()

    def test_get_builders(self):
        """Test getting the builders."""
        builders = self.factory.get_builders()
        self.assertIsInstance(builders["gas_species"], GasSpeciesBuilder)
        self.assertIsInstance(
            builders["preset_gas_species"], PresetGasSpeciesBuilder
        )

    def test_get_strategy_gas_species(self):
        """Test getting a gas species strategy."""
        parameters = {
            "name": "Oxygen",
            "molar_mass": 0.032,
            "vapor_pressure_strategy": ConstantVaporPressureStrategy(50),
            "condensable": False,
            "concentration": 1.2,
        }
        gas_species = self.factory.get_strategy("gas_species", parameters)
        self.assertIsInstance(gas_species, GasSpecies)
        self.assertEqual(gas_species.name, parameters["name"])
        self.assertEqual(gas_species.molar_mass, parameters["molar_mass"])
        self.assertEqual(
            gas_species.pure_vapor_pressure_strategy,
            parameters["vapor_pressure_strategy"],
        )
        self.assertEqual(gas_species.condensable, parameters["condensable"])
        self.assertEqual(gas_species.concentration, parameters["concentration"])
        self.assertIsInstance(gas_species, GasSpecies)

    def test_get_strategy_preset_gas_species(self):
        """Test getting a preset gas species strategy."""
        parameters = {}
        preset_gas_species = self.factory.get_strategy(
            "preset_gas_species", parameters
        )
        self.assertIsInstance(preset_gas_species, GasSpecies)
        self.assertEqual(preset_gas_species.molar_mass, 0.1)
        self.assertIsInstance(
            preset_gas_species.pure_vapor_pressure_strategy,
            ConstantVaporPressureStrategy,
        )
        self.assertEqual(preset_gas_species.condensable, False)
        self.assertEqual(preset_gas_species.concentration, 1.0)

    def test_get_strategy_invalid(self):
        """Test getting an invalid strategy."""
        with self.assertRaises(ValueError):
            self.factory.get_strategy("invalid_type", {})



"""Test the gas species builder, for building gas species objects.
and checking/validation of the parameters.
"""

import pytest
from particula.gas.species import GasSpecies
from particula.gas.species_builders import (
    GasSpeciesBuilder,
    PresetGasSpeciesBuilder,
)
from particula.gas.vapor_pressure_strategies import (
    ConstantVaporPressureStrategy,
)


def test_set_name():
    """Test setting the name of the gas species."""
    builder = GasSpeciesBuilder()
    builder.set_name("Oxygen")
    assert builder.name == "Oxygen"


def test_set_molar_mass():
    """Test setting the molar mass of the gas species."""
    builder = GasSpeciesBuilder()
    builder.set_molar_mass(0.032)
    assert builder.molar_mass == 0.032

    with pytest.raises(ValueError):
        builder.set_molar_mass(-0.032)


def test_set_vapor_pressure_strategy():
    """Test setting the vapor pressure strategy of the gas species."""
    builder = GasSpeciesBuilder()
    strategy = ConstantVaporPressureStrategy(0.0)
    builder.set_vapor_pressure_strategy(strategy)
    assert builder.vapor_pressure_strategy == strategy


def test_set_condensable():
    """Test setting the condensable bool of the gas species."""
    builder = GasSpeciesBuilder()
    builder.set_condensable(True)
    assert builder.condensable is True

    builder.set_condensable(False)
    assert builder.condensable is False


def test_set_concentration():
    """Test setting the concentration of the gas species."""
    builder = GasSpeciesBuilder()
    builder.set_concentration(1.0)
    assert builder.concentration == 1.0

    with pytest.raises(ValueError):
        builder.set_concentration(-1.0)

    builder.set_concentration(1e-3, concentration_units="kg/m^3")
    assert builder.concentration == 1.0e-3


def test_set_parameters():
    """Test setting all parameters at once."""
    builder = GasSpeciesBuilder()
    vapor_obj = ConstantVaporPressureStrategy(0.0)
    parameters = {
        "name": "Oxygen",
        "molar_mass": 0.032,
        "vapor_pressure_strategy": vapor_obj,
        "condensable": True,
        "concentration": 1.0,
    }
    builder.set_parameters(parameters)
    assert builder.name == "Oxygen"
    assert builder.molar_mass == 0.032
    assert builder.vapor_pressure_strategy == vapor_obj
    assert builder.condensable is True
    assert builder.concentration == 1.0


def test_missing_required_parameter():
    """Test missing a required parameter in the parameters dict."""
    builder = GasSpeciesBuilder()
    parameters = {
        "name": "Oxygen",
        "molar_mass": 0.032,
        "vapor_pressure_strategy": ConstantVaporPressureStrategy(0.0),
        "condensable": True,
    }
    with pytest.raises(ValueError):
        builder.set_parameters(parameters)


def test_invalid_parameter():
    """Test an invalid parameter in the parameters dict."""
    builder = GasSpeciesBuilder()
    parameters = {
        "name": "Oxygen",
        "molar_mass": 0.032,
        "vapor_pressure_strategy": ConstantVaporPressureStrategy(0.0),
        "condensable": True,
        "concentration": 1.0,
        "invalid_param": 123,
    }
    with pytest.raises(ValueError):
        builder.set_parameters(parameters)


def test_build():
    """Test building the gas species object."""
    builder = GasSpeciesBuilder()
    parameters = {
        "name": "Oxygen",
        "molar_mass": 0.032,
        "vapor_pressure_strategy": ConstantVaporPressureStrategy(0.0),
        "condensable": True,
        "concentration": 1.0,
    }
    builder.set_parameters(parameters)
    gas_species = builder.build()
    assert isinstance(gas_species, GasSpecies)
    assert gas_species.get_name() == "Oxygen"


def test_preset_builder():
    """Test the preset gas species builder."""
    builder = PresetGasSpeciesBuilder()
    gas_species = builder.build()
    assert isinstance(gas_species, GasSpecies)
    assert gas_species.get_name() == "Preset100"
    assert gas_species.get_molar_mass() == 0.100
    assert gas_species.get_condensable() is False
    assert gas_species.get_concentration() == 1.0



"""
Test for convert_functional_group.py
"""

from particula.activity.convert_functional_group import (
    convert_to_oh_equivalent
)


def test_convert_to_oh_eqivalent():
    """test for convert_to_oh_equivalent function."""

    molar_mass_ratio = 18.016 / 250
    oxygen2carbon = 0.3

    new_oxygen2carbon, new_molar_mass_ratio = (
        convert_to_oh_equivalent(
            oxygen2carbon=oxygen2carbon,
            molar_mass_ratio=molar_mass_ratio,
            functional_group=None,
        )
    )
    assert new_oxygen2carbon == oxygen2carbon
    assert new_molar_mass_ratio == molar_mass_ratio



"""Test the vapor transition correction properties."""

import numpy as np
from particula.particles.properties import vapor_transition_correction


def test_vapor_transition_correction_scalar():
    """Test the vapor_transition_correction function with scalar values."""
    knudsen_number = 0.5
    mass_accommodation = 0.1
    result = vapor_transition_correction(knudsen_number, mass_accommodation)
    expected = (0.75 * 0.1 * (1 + 0.5)) / (
        (0.5**2 + 0.5) + 0.283 * 0.1 * 0.5 + 0.75 * 0.1
    )
    assert np.isclose(result, expected), "Failed for scalar inputs"


def test_vapor_transition_correction_array():
    """Test the vapor_transition_correction function with NumPy arrays."""
    knudsen_number = np.array([0.5, 1.0, 2.0])
    mass_accommodation = np.array([0.1, 0.2, 0.3])
    result = vapor_transition_correction(knudsen_number, mass_accommodation)
    expected = (0.75 * mass_accommodation * (1 + knudsen_number)) / (
        (knudsen_number**2 + knudsen_number)
        + 0.283 * mass_accommodation * knudsen_number
        + 0.75 * mass_accommodation
    )
    assert np.allclose(result, expected), "Failed for array inputs"


def test_vapor_transition_correction_zero_knudsen():
    """Test the vapor_transition_correction function with
    zero Knudsen number."""
    knudsen_number = 0
    mass_accommodation = 0.2
    result = vapor_transition_correction(knudsen_number, mass_accommodation)
    expected = (0.75 * 0.2 * (1 + 0)) / (
        (0**2 + 0) + 0.283 * 0.2 * 0 + 0.75 * 0.2
    )
    assert np.isclose(result, expected), "Failed for zero Knudsen number"


def test_vapor_transition_correction_zero_accommodation():
    """Test the vapor_transition_correction function with
    zero mass accommodation."""
    knudsen_number = 1
    mass_accommodation = 0
    result = vapor_transition_correction(knudsen_number, mass_accommodation)
    expected = (0.75 * 0 * (1 + 1)) / ((1**2 + 1) + 0.283 * 0 * 1 + 0.75 * 0)
    assert np.isclose(result, expected), "Failed for zero mass accommodation"


def test_vapor_transition_correction_high_values():
    """Test the vapor_transition_correction function with high values."""
    knudsen_number = 100
    mass_accommodation = 1
    result = vapor_transition_correction(knudsen_number, mass_accommodation)
    expected = (0.75 * 1 * (1 + 100)) / (
        (100**2 + 100) + 0.283 * 1 * 100 + 0.75 * 1
    )
    assert np.isclose(result, expected), "Failed for high values"



"""conversion functions common for aerosol processing
"""

from typing import Union, Tuple, Any, List, Dict
from numpy.typing import NDArray
import numpy as np


def coerce_type(data, dtype):
    """
    Coerces data to dtype if it is not already of that type.

    Example:
        >>> coerce_type(1, float)
        1.0
        >>> coerce_type([1, 2, 3], np.ndarray)
        array([1, 2, 3])

    Args:
        data: The data to be coerced.
        dtype: The desired data type.

    Returns:
        The coerced data.

    Raises:
        ValueError: If the data cannot be coerced to the desired type.
    """
    if not isinstance(data, dtype):
        try:
            return np.array(data) if dtype == np.ndarray else dtype(data)
        except (ValueError, TypeError) as exc:
            raise ValueError(f"Could not coerce {data} to {dtype}") from exc
    return data


def round_arbitrary(
    values: Union[float, list[float], np.ndarray],
    base: Union[float, np.float64] = 1.0,
    mode: str = "round",
    nonzero_edge: bool = False,
) -> Union[float, NDArray[np.float64]]:
    """
    Rounds the input values to the nearest multiple of the base.

    For values exactly halfway between rounded decimal values, "Bankers
    rounding applies" rounds to the nearest even value. Thus 1.5 and 2.5
    round to 2.0, -0.5 and 0.5 round to 0.0, etc.

    Args:
        values: The values to be rounded.
        base: The base to which the values should be rounded.
        mode: The rounding mode: 'round', 'floor', 'ceil'
        nonzero_edge: If true the zero values are replaced
        by the original values.

    Returns:
        rounded: The rounded values.
    """
    # Check if values is a NumPy array
    working_values = coerce_type(values, np.ndarray)
    base = coerce_type(base, float)

    # Validate base parameter
    if not isinstance(base, float) or base <= 0:
        raise ValueError("base must be a positive float")
    # Validate mode parameter
    if mode not in ["round", "floor", "ceil"]:
        raise ValueError("mode must be one of ['round', 'floor', 'ceil']")

    # Calculate rounding factors
    factor = np.array([-0.5, 0, 0.5])

    # Compute rounded values
    rounded = base * np.round(
        working_values / base
        + factor[np.array(["floor", "round", "ceil"]).tolist().index(mode)]
    )

    # Apply round_nonzero mode
    if nonzero_edge:
        rounded = np.where(rounded != 0, rounded, working_values)

    return float(rounded) if isinstance(values, float) else rounded


def radius_diameter(value: float, to_diameter: bool = True) -> float:
    """
    Convert a radius to a diameter, or vice versa.

    Args:
        value: The value to be converted.
        to_diameter: If True, convert from radius to diameter.
        If False, convert from diameter to radius.

    Returns:
        The converted value.
    """
    return value * 2 if to_diameter else value / 2


def volume_to_length(
    volume: Union[float, NDArray[np.float64]], length_type: str = "radius"
) -> Union[float, NDArray[np.float64]]:
    """
    Convert a volume to a radius or diameter.

    Args:
        volume: The volume to be converted.
        length_type: The type of length to convert to ('radius' or 'diameter')
        Default is 'radius'.

    Returns:
        The converted length.
    """

    if length_type not in ["radius", "diameter"]:
        raise ValueError("length_type must be radius or diameter")

    radius = (volume * 3 / (4 * np.pi)) ** (1 / 3)

    return radius if length_type == "radius" else radius * 2


def length_to_volume(
    length: Union[float, np.ndarray], length_type: str = "radius"
) -> Union[float, np.ndarray]:
    """
    Convert radius or diameter to volume.

    Args:

        length: The length to be converted.
        length_type: The type of length ('radius' or 'diameter').
            Default is 'radius'.

    Returns:
    --------
        The volume.
    """
    if length_type == "diameter":
        length = length / 2
    elif length_type != "radius":
        raise ValueError("length_type must be radius or diameter")
    return (4 / 3) * np.pi * (length**3)


def kappa_volume_solute(
    volume_total: Union[float, np.ndarray],
    kappa: Union[float, np.ndarray],
    water_activity: Union[float, np.ndarray],
) -> Union[float, np.ndarray]:
    """
    Calculate the volume of solute in a volume of total solution,
    given the kappa parameter and water activity.

    Args:

        volume_total: The volume of the total solution.
        kappa: The kappa parameter.
        water_activity: The water activity.

    Returns:
    --------
        The volume of solute as a numpy array.
    """

    kappa = max(kappa, 1e-16)  # Avoid division by zero
    if water_activity <= 1e-16:  # early return for low water activity
        return volume_total

    vol_factor = (water_activity - 1) / (
        water_activity * (1 - kappa - 1 / water_activity)
    )
    return volume_total * np.array(vol_factor)


def kappa_volume_water(
    volume_solute: Union[float, NDArray[np.float64]],
    kappa: Union[float, NDArray[np.float64]],
    water_activity: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """
    Calculate the volume of water given volume of solute, kappa parameter,
    and water activity.

    Args:

        volume_solute: The volume of solute.
        kappa: The kappa parameter.
        water_activity: The water activity.

    Returns:
    --------
        The volume of water as a float.
    """
    # Avoid division by zero
    water_activity = min(water_activity, 1 - 1e-16)

    if water_activity <= 1e-16:  # early return for low water activity
        return volume_solute * 0

    return volume_solute * kappa / (1 / water_activity - 1)


def kappa_from_volume(
    volume_solute: Union[float, np.ndarray],
    volume_water: Union[float, np.ndarray],
    water_activity: Union[float, np.ndarray],
) -> Union[float, np.ndarray]:
    """
    Calculate the kappa parameter from the volume of solute and water,
    given the water activity.

    Args:

        volume_solute: The volume of solute.
        volume_water: The volume of water.
        water_activity: The water activity.

    Returns:
    --------
        The kappa parameter as a float.
    """
    # Avoid division by zero
    water_activity = np.where(
        water_activity > 1 - 1e-16, 1 - 1e-16, water_activity
    )

    return (1 / water_activity - 1) * volume_water / volume_solute


def mole_fraction_to_mass_fraction(
    mole_fraction0: float, molecular_weight0: float, molecular_weight1: float
) -> Tuple[float, float]:
    """
    Convert mole fraction to mass fraction.

    Args:
        mole_fraction0: The mole fraction of the first component.
        molecular_weight0: The molecular weight of the first component.
        molecular_weight1: The molecular weight of the second component.

    Returns:
        A tuple containing the mass fractions of the two components as floats.
    """
    mass_fraction0 = (
        mole_fraction0
        * molecular_weight0
        / (
            mole_fraction0 * molecular_weight0
            + (1 - mole_fraction0) * molecular_weight1
        )
    )
    mass_fraction1 = 1 - mass_fraction0
    return mass_fraction0, mass_fraction1


def mole_fraction_to_mass_fraction_multi(
    mole_fractions: list[float], molecular_weights: list[float]
) -> list[float]:
    """Convert mole fractions to mass fractions for N components.
    Assumes that sum(mole_fractions) == 1.

    Args:
        mole_fractions: A list of mole fractions.
        molecular_weights: A list of molecular weights.

    Returns:
        A list of mass fractions.
    """
    if np.sum(mole_fractions) != 1:
        raise ValueError("Sum of mole fractions must be 1")

    total_molecular_weight = np.sum(
        [mf * mw for mf, mw in zip(mole_fractions, molecular_weights)]
    )
    return [
        mf * mw / total_molecular_weight
        for mf, mw in zip(mole_fractions, molecular_weights)
    ]


def mass_concentration_to_mole_fraction(
    mass_concentrations: NDArray[np.float64], molar_masses: NDArray[np.float64]
) -> NDArray[np.float64]:
    """
    Convert mass concentrations to mole fractions for N components.

    Args:
        mass_concentrations: A list or ndarray of mass concentrations
        (e.g., kg/m^3).
        molar_masses: A list or ndarray of molecular weights (e.g., g/mol).

    Returns:
        An ndarray of mole fractions.

    Note:
        The mole fraction of a component is given by the ratio of its molar
        concentration to the total molar concentration of all components.
    """
    # Convert mass concentrations to moles for each component
    moles = mass_concentrations / molar_masses

    # Calculate total moles in the mixture
    total_moles = np.sum(moles)

    return moles / total_moles


def mass_concentration_to_volume_fraction(
    mass_concentrations: NDArray[np.float64], densities: NDArray[np.float64]
) -> NDArray[np.float64]:
    """Convert mass concentrations to volume fractions for N components.

    Args:
        mass_concentrations: A list or ndarray of mass concentrations
        (e.g., kg/m^3).
        densities: A list or ndarray of densities of each component
        (e.g., kg/m^3).

    Returns:
        An ndarray of volume fractions.

    Note:
        The volume fraction of a component is calculated by dividing the volume
        of that component (derived from mass concentration and density) by the
        total volume of all components.
    """
    # Calculate volumes for each component using mass concentration and density
    volumes = mass_concentrations / densities  # [:, np.newaxis]

    # Calculate total volume of the mixture
    total_volume = np.sum(volumes)  # axis=0

    # Calculate volume fractions by dividing the volume of each component by
    # the total volume

    return volumes / total_volume


def mass_fraction_to_volume_fraction(
    mass_fraction: float, density_solute: float, density_solvent: float
) -> Tuple[float, float]:
    """
    Converts the mass fraction of a solute to the volume fraction in a
    binary mixture.

    Args:
        mass_fraction (float): The mass fraction of the solute in the mixture.
        density_solute (float): The density of the solute.
        density_solvent (float): The density of the solvent.

    Returns:
        Tuple[float, float]: A tuple containing the volume fraction of the
            solute and solvent in the mixture.

    Example:
        If `mass_fraction` is 0.5, `density_solute` is 1.5 g/cm^3, and
        `density_solvent` is 2 g/cm^3, this function returns (0.5714, 0.4285),
        indicating that the solute and solvent occupy 57% and 42% of the
        mixture's volume, respectively.
    """
    volume_fraction_solute = (
        mass_fraction
        / density_solute
        / (
            mass_fraction / density_solute
            + (1 - mass_fraction) / density_solvent
        )
    )
    volume_fraction_solvent = 1 - volume_fraction_solute
    return volume_fraction_solute, volume_fraction_solvent


def volume_water_from_volume_fraction(
    volume_solute_dry: float, volume_fraction_water: float
) -> float:
    """
    Calculates the volume of water in a volume of solute, given the volume
    fraction of water in the mixture.

    Args:
        volume_solute_dry (float): The volume of the solute, excluding water.
        volume_fraction_water (float): The volume fraction of water in the
                            mixture, expressed as a decimal between 0 and 1.

    Returns:
        float: The volume of water in the mixture, in the same units as
            `volume_solute_dry`.

    Example:
        If `volume_solute_dry` is 100 mL and `volume_fraction_water` is 0.8,
        this function returns 400 mL, indicating that there are 400 mL of water
        in the total 100 mL + 400 mL mixture.
    """
    return (
        volume_fraction_water * volume_solute_dry / (1 - volume_fraction_water)
    )


def effective_refractive_index(
    m_zero: Union[float, complex],
    m_one: Union[float, complex],
    volume_zero: float,
    volume_one: float,
) -> Union[float, complex]:
    """
    Calculate the effective refractive index of a mixture of two solutes, given
    the refractive index of each solute and the volume of each solute. The
    mixing is based on volume-weighted molar refraction.

    Args:
        m_zero: The refractive index of solute 0.
        m_one: The refractive index of solute 1.
        volume_zero: The volume of solute 0.
        volume_one: The volume of solute 1.

    Returns:
        The effective refractive index of the mixture.

    References:
        Liu, Y., &#38; Daum, P. H. (2008). Relationship of refractive
        index to mass density and self-consistency mixing rules for
        multicomponent mixtures like ambient aerosols.
        Journal of Aerosol Science, 39(11), 974-986.
        https://doi.org/10.1016/j.jaerosci.2008.06.006
    """
    volume_total = volume_zero + volume_one
    r_effective = volume_zero / volume_total * (m_zero - 1) / (
        m_zero + 2
    ) + volume_one / volume_total * (m_one - 1) / (
        m_one + 2
    )  # molar refraction mixing

    # convert to refractive index
    return (2 * r_effective + 1) / (1 - r_effective)


def convert_sizer_dn(
    diameter: np.ndarray, dn_dlogdp: np.ndarray, inverse: bool = False
) -> np.ndarray:
    """
    Converts the sizer data from dn/dlogdp to d_num.

    The bin width is defined as the  difference between the upper and lower
    diameter limits of each bin. This function calculates the bin widths
    based on the input diameter array. Assumes a log10 scale for dp edges.

    Args:
        diameter (np.ndarray): Array of particle diameters.
        dn_dlogdp (np.ndarray): Array of number concentration of particles per
        unit logarithmic diameter.
        inverse (bool): If True, converts from d_num to dn/dlogdp.

    Returns:
        np.ndarray: Array of number concentration of particles
        per unit diameter.

    References:
        Eq: dN/dlogD_p = dN/( log(D_{p-upper}) - log(D_{p-lower}) )
        https://tsi.com/getmedia/1621329b-f410-4dce-992b-e21e1584481a/
        PR-001-RevA_Aerosol-Statistics-AppNote?ext=.pdf
    """
    assert len(diameter) > 0, "Inputs must be non-empty arrays."
    # Compute the bin widths
    delta = np.zeros_like(diameter)
    delta[:-1] = np.diff(diameter)
    delta[-1] = delta[-2] ** 2 / delta[-3]

    # Compute the lower and upper bin edges
    lower = diameter - delta / 2
    upper = diameter + delta / 2

    # if dn_dlogdp.ndim == 2:
    #     # expand diameter by one dimension so it can be broadcast
    #     lower = np.expand_dims(lower, axis=1)
    #     upper = np.expand_dims(upper, axis=1)
    if inverse:
        # Convert from dn to dn/dlogdp
        return dn_dlogdp / np.log10(upper / lower)

    return dn_dlogdp * np.log10(upper / lower)


def list_to_dict(list_of_str: list) -> dict:
    """
    Converts a list of strings to a dictionary. The keys are the strings
    and the values are the index of the string in the list.

    Args:
        list_of_str (list): A non-empty list of strings.

    Returns:
        dict: A dictionary where the keys are the strings and the values are
            the index of the string in the list.
    """
    assert all(list_of_str), "Input list_of_str must not be empty."
    assert all(
        isinstance(item, str) for item in list_of_str
    ), "Input \
        list_of_str must contain only strings."

    # Create a dictionary from the list of strings using a dictionary
    # comprehension
    return {str_val: i for i, str_val in enumerate(list_of_str)}


def get_values_in_dict(
    key_list: List[str], dict_to_check: Dict[str, Any]
) -> List[Any]:
    """
    Returns a list of values for keys in a dictionary.

    Example:
    >>> my_dict = {'a': 1, 'b': 2, 'c': 3}
    >>> get_values_in_dict(['a', 'c'], my_dict)
    [1, 3]

    Args:
        key_list: List of keys to check in the dictionary.
        dict_to_check: The dictionary to check for the given keys.

    Returns:
        List: A list of values for keys in the dictionary.

    Raises:
        KeyError: If any of the keys in the `key_list` are not present in
        the dictionary.
    """
    values = []
    for key in key_list:
        if key in dict_to_check:
            values.append(dict_to_check[key])
        else:
            raise KeyError(
                f"Key '{key}' not found in the dictionary. Available keys:"
                + f"{list(dict_to_check.keys())}"
            )
    return values


def data_shape_check(
    time: np.ndarray,
    data: np.ndarray,
    header: list,
) -> np.ndarray:
    """
    Check the shape of the input data and header list, and reshape the data if
    necessary. The data array can be 1D or 2D. If the data array is 2D, the
    time array must match the last dimensions of the data array. If the data
    array is 1D, the header list must be a single entry.

    Args:
        time (np.ndarray): 1D array of time values.
        data (np.ndarray): 1D or 2D array of data values.
        header (list): List of header values.

    Returns:
        Reshaped data array.

    Raises:
        ValueError: If the length of the header list does not match the first
        dimension of the data array.
    """

    # Check if data_new is 2D or 1D
    if len(data.shape) == 2:
        # Check if time matches the dimensions of data
        if len(time) == data.shape[0] and len(time) == data.shape[1]:
            concatenate_axis_new = 1  # Default to the axis=1
        else:
            # Find the axis that doesn't match the length of time
            concatenate_axis_new = np.argwhere(
                np.array(data.shape) != len(time)
            ).flatten()[0]
        # Reshape new data so the concatenate axis is axis=1
        data = np.moveaxis(data, concatenate_axis_new, 1)

        # check header list length matches data_new shape
        if len(header) != data.shape[1]:
            print(
                f"header len: {len(header)} vs. data.shape: \
                  {data.shape}"
            )
            print(header)
            raise ValueError(
                "Header list length must match the second \
                              dimension of data_new."
            )
    elif len(header) == 1:
        # Reshape new data so the concatenate axis is axis=1
        data = np.expand_dims(data, 1)

    else:
        raise ValueError(
            "Header list must be a single entry if data_new \
                              is 1D."
        )
    return data


def distribution_convert_pdf_pms(
    x_array: np.ndarray, distribution: np.ndarray, to_pdf: bool = True
) -> np.ndarray:
    """
    Convert between a probability density function (PDF) and a probability
    mass spectrum (PMS) based on the specified direction.

    Args:
        x_array : An array of radii corresponding to the bins of the
            distribution, shape (m).
        distribution : The concentration values of the distribution
            (either PDF or PMS) at the given radii. Supports broadcasting
            across x_array (n,m).
        to_PDF : Direction of conversion. If True, converts PMS to PDF.
            If False, converts PDF to PMS.

    Returns:
        converted_distribution : The converted distribution array
            (either PDF or PMS).
    """

    # Calculate the differences between consecutive x_array values for bin
    # widths.
    delta_x_array = np.empty_like(x_array)
    delta_x_array[:-1] = np.diff(x_array)
    # For the last bin, extrapolate the width assuming constant growth
    # rate from the last two bins.
    delta_x_array[-1] = delta_x_array[-2] ** 2 / delta_x_array[-3]

    # Converting PMS to PDF by dividing the PMS values by the bin widths. or
    # Converting PDF to PMS by multiplying the PDF values by the bin widths.
    return (
        distribution / delta_x_array
        if to_pdf
        else distribution * delta_x_array
    )



"""
Tests for the sigma_relative_velocity_variance module.
"""

import pytest
import numpy as np

from particula.dynamics.coagulation.turbulent_dns_kernel.sigma_relative_velocity_ao2008 import (
    get_relative_velocity_variance,
)


def test_get_relative_velocity_variance_array():
    """
    Test get_relative_velocity_variance with NumPy array inputs.
    """

    expected_shape = (3, 3)
    result = get_relative_velocity_variance(
        fluid_rms_velocity=0.05,
        collisional_radius=np.array([0.05, 0.1, 0.2]),
        particle_inertia_time=np.array([2, 3, 5]) * 1e-6,
        particle_velocity=np.array([0.3, 0.4, 0.5]),
        taylor_microscale=0.05,
        eulerian_integral_length=2.0,
        lagrangian_integral_time=0.3,
    )

    assert (
        result.shape == expected_shape
    )


def test_invalid_inputs():
    """
    Ensure validation errors are raised for invalid inputs.
    """
    with pytest.raises(ValueError):
        get_relative_velocity_variance(
            -0.5, 0.05, 0.02, 0.3, 0.05, 1.0, 0.2
        )  # Negative fluid_rms_velocity

    with pytest.raises(ValueError):
        get_relative_velocity_variance(
            0.5, -0.05, 0.02, 0.3, 0.05, 1.0, 0.2
        )  # Negative collisional_radius

    with pytest.raises(ValueError):
        get_relative_velocity_variance(
            0.5, 0.05, -0.02, 0.3, 0.05, 1.0, 0.2
        )  # Negative particle_inertia_time

    with pytest.raises(ValueError):
        get_relative_velocity_variance(
            0.5, 0.05, 0.02, -0.3, 0.05, 1.0, 0.2
        )  # Negative particle_velocity


def test_edge_cases():
    """
    Test compute_relative_velocity_variance with extreme values.
    """
    fluid_rms_velocity = np.array(
        [1e-6, 1e-3, 10.0]
    )  # Very small and large values
    collisional_radius = np.array(
        [1e-6, 1e-3, 10.0]
    )  # Very small and large values
    particle_inertia_time = np.array(
        [1e-6, 1e-3, 10.0]
    )  # Very small and large values
    particle_velocity = np.array(
        [1e-6, 1e-3, 10.0]
    )  # Very small and large values

    result = get_relative_velocity_variance(
        fluid_rms_velocity=fluid_rms_velocity,
        collisional_radius=collisional_radius,
        particle_inertia_time=particle_inertia_time,
        particle_velocity=particle_velocity,
        taylor_microscale=0.05,
        eulerian_integral_length=1.0,
        lagrangian_integral_time=0.2,
    )

    assert np.all(np.isfinite(result)), "Expected all values to be finite"



"""Test builders for activity strategies
for error and validation handling.

Correctness is tested by activity_strategies_test.py.
"""

import pytest
import numpy as np
from particula.particles.activity_builders import (
    ActivityIdealMolarBuilder,
    ActivityIdealMassBuilder,
    ActivityKappaParameterBuilder,
)


def test_build_ideal_activity_mass():
    """Test building an IdealActivityMass object."""
    builder = ActivityIdealMassBuilder()
    activity = builder.build()
    assert activity.get_name() == "ActivityIdealMass"


def test_build_ideal_activity_molar_parameter():
    """Test that providing a negative molar mass raises a ValueError."""
    builder = ActivityIdealMolarBuilder()
    with pytest.raises(ValueError) as excinfo:
        builder.set_molar_mass(-1)
    assert "Argument 'molar_mass' must be positive." in str(excinfo.value)

    # test positive molar mass
    builder.set_molar_mass(1)
    assert builder.molar_mass == 1

    # test array of molar masses
    builder.set_molar_mass(np.array([1, 2, 3]))
    np.testing.assert_array_equal(builder.molar_mass, np.array([1, 2, 3]))

    # test setting molar mass units
    builder.set_molar_mass(1e-3, molar_mass_units="kg/mol")
    assert builder.molar_mass == 1e-3

    # test setting molar mass units for array
    builder.set_molar_mass(np.array([1, 2, 3])/1000, molar_mass_units="kg/mol")
    np.testing.assert_array_equal(
        builder.molar_mass, np.array([1e-3, 2e-3, 3e-3])
    )


def test_build_ideal_activity_molar_dict():
    """Test building an IdealActivityMolar object."""
    builder_dict = ActivityIdealMolarBuilder()
    parameters = {"molar_mass": 1, "molar_mass_units": "kg/mol"}
    builder_dict.set_parameters(parameters)
    assert builder_dict.molar_mass == 1

    # build the object
    activity = builder_dict.build()
    assert activity.__class__.__name__ == "ActivityIdealMolar"


def test_build_ideal_activity_molar_missing_parameters():
    """Test building an IdealActivityMolar object with missing parameters."""
    builder_missing = ActivityIdealMolarBuilder()
    with pytest.raises(ValueError) as excinfo:
        builder_missing.build()
    assert "Required parameter(s) not set: molar_mass" in str(excinfo.value)


def test_build_kappa_parameter_activity_set_kappa():
    """Testing setting kappa parameter."""
    builder = ActivityKappaParameterBuilder()
    builder.set_kappa(1)
    assert builder.kappa == 1

    # test setting kappa units
    builder.set_kappa(1)
    assert builder.kappa == 1

    # test setting kappa units for array
    builder.set_kappa(np.array([1, 2, 3]))
    np.testing.assert_array_equal(builder.kappa, np.array([1, 2, 3]))

    # test setting kappa units, have no effect
    builder.set_kappa(1, kappa_units="no_units")
    assert builder.kappa == 1

    # test negative kappa
    with pytest.raises(ValueError) as excinfo:
        builder.set_kappa(-1)
    assert "Kappa parameter must be a positive value." in str(excinfo.value)


def test_build_kappa_parameter_activity_set_density():
    """Testing setting density parameter."""
    builder = ActivityKappaParameterBuilder()
    builder.set_density(1, "kg/m^3")
    assert builder.density == pytest.approx(1, rel=1e-5)

    # test setting density units
    builder.set_density(1e3, density_units="kg/m^3")
    assert builder.density == pytest.approx(1e3, rel=1e-5)

    # test setting density units for array
    builder.set_density(np.array([1, 2, 3])*1e3, density_units="kg/m^3")
    np.testing.assert_allclose(
        builder.density, np.array([1e3, 2e3, 3e3]), atol=1e-5
    )

    # test negative density
    with pytest.raises(ValueError) as excinfo:
        builder.set_density(-1, "kg/m^3")
    assert "Argument 'density' must be positive." in str(excinfo.value)


def test_build_kappa_parameter_activity_set_molar_mass():
    """Testing setting molar mass parameter."""
    builder = ActivityKappaParameterBuilder()
    builder.set_molar_mass(1)
    assert builder.molar_mass == 1

    # test setting molar mass units
    builder.set_molar_mass(0.001, molar_mass_units="kg/mol")
    assert builder.molar_mass == pytest.approx(1e-3, rel=1e-5)

    # test setting molar mass units for array
    builder.set_molar_mass(np.array([1, 2, 3])/1000, molar_mass_units="kg/mol")
    np.testing.assert_allclose(
        builder.molar_mass, np.array([1e-3, 2e-3, 3e-3]), atol=1e-5
    )

    # test negative molar mass
    with pytest.raises(ValueError) as excinfo:
        builder.set_molar_mass(-1)
    assert "Argument 'molar_mass' must be positive." in str(excinfo.value)


def test_build_kappa_parameter_activity_set_water_index():
    """Testing setting water index"""
    builder = ActivityKappaParameterBuilder()
    builder.set_water_index(1)
    assert builder.water_index == 1

    # test ignore units
    builder.set_water_index(1, water_index_units="no_units")
    assert builder.water_index == 1


def test_build_kappa_parameter_activity_dict():
    """Test building a KappaParameterActivity object."""
    builder_dict = ActivityKappaParameterBuilder()
    parameters = {
        "kappa": np.array([1, 2, 3]),
        "density": np.array([1, 2, 3]),
        "density_units": "kg/m^3",
        "molar_mass": np.array([1, 2, 3]),
        "water_index": 1,
    }
    builder_dict.set_parameters(parameters)
    np.testing.assert_allclose(
        builder_dict.kappa, parameters["kappa"], atol=1e-5
    )
    np.testing.assert_allclose(
        builder_dict.density, parameters["density"], atol=1e-5
    )
    np.testing.assert_allclose(
        builder_dict.molar_mass, parameters["molar_mass"], atol=1e-5
    )
    assert builder_dict.water_index == 1

    # build the object
    activity = builder_dict.build()
    assert activity.__class__.__name__ == "ActivityKappaParameter"

    # test missing parameters
    builder_missing2 = ActivityKappaParameterBuilder()
    parameters = {
        "kappa": np.array([0, 1]),
        "density": np.array([1, 1]),
        "molar_mass": np.array([1, 2]),
    }
    with pytest.raises(ValueError) as excinfo:
        builder_missing2.set_parameters(parameters)
    assert "Missing required parameter(s): water_index" in str(excinfo.value)



"""Tests for the logging setup module.

This module verifies that the logger configuration is correctly loaded,
including the correct setup of handlers and formatters as specified in
the configuration dictionary.
"""

import logging
import logging.handlers

logger = logging.getLogger("particula")


def test_setup_configures_logger():
    """Test that the logger name is set to 'particula'."""
    # Arrange
    expected_logger_name = "particula"

    # Assert
    assert logger.name == expected_logger_name

    logger.info("Logger Test: This is a test message.")
    logger.debug("Logger Test: This is a debug message.")
    logger.error("Logger Test: This is an error message.")
    logger.warning("Logger Test: This is a warning message.")
    logger.critical("Logger Test: This is a critical message.")
    logger.exception("Logger Test: This is an exception message.")



"""Condensation dynamics exposed via __init__.

If you want specific sub functions, import them directly from
`particula.next.dynamics.condensation.xxx.xxx` instead.
"""

# pylint: disable=unused-import, disable=line-too-long
# flake8: noqa
# pyright: basic

from particula.dynamics import condensation
from particula.dynamics import coagulation
from particula.dynamics import properties

from particula.dynamics.dilution import (
    dilution_rate,
    volume_dilution_coefficient,
)

from particula.dynamics.wall_loss import (
    rectangle_wall_loss_rate,
    spherical_wall_loss_rate,
)

from particula.dynamics.particle_process import (
    MassCondensation,
    Coagulation,
)



"""
Blending weights for the BAT model.
"""

import numpy as np
import pytest
from particula.activity.bat_blending import bat_blending_weights


def test_bat_blending_weights():
    """test for bat_blending_weights function."""
    molar_mass_ratio = 18.016 / 250
    oxygen2carbon = 0.3

    weights = bat_blending_weights(
        molar_mass_ratio=molar_mass_ratio, oxygen2carbon=oxygen2carbon
    )
    assert np.all(weights >= 0)
    assert np.all(weights <= 1)


def test_bat_blending_weights_float():
    """Test bat_blending_weights with float inputs."""
    molar_mass_ratio = 0.5
    oxygen2carbon = 1.0
    result = bat_blending_weights(molar_mass_ratio, oxygen2carbon)
    assert result.shape == (3,)
    assert np.all(result >= 0) and np.all(result <= 1)


def test_bat_blending_weights_array():
    """Test bat_blending_weights with array inputs."""
    molar_mass_ratio = np.array([0.5, 0.6])
    oxygen2carbon = np.array([1.0, 1.2])
    result = bat_blending_weights(molar_mass_ratio, oxygen2carbon)
    assert result.shape == (2, 3)
    assert np.all(result >= 0) and np.all(result <= 1)


def test_bat_blending_weights_edge_cases():
    """Test bat_blending_weights with edge cases."""
    molar_mass_ratio = 0.5
    oxygen2carbon = 0.0
    result = bat_blending_weights(molar_mass_ratio, oxygen2carbon)
    assert result.shape == (3,)
    assert np.all(result >= 0) and np.all(result <= 1)

    oxygen2carbon = 10.0
    result = bat_blending_weights(molar_mass_ratio, oxygen2carbon)
    assert result.shape == (3,)
    assert np.all(result >= 0) and np.all(result <= 1)


def test_bat_blending_weights_invalid_inputs():
    """Test bat_blending_weights with invalid inputs."""
    molar_mass_ratio = -0.5
    oxygen2carbon = 1.0
    with pytest.raises(ValueError):
        bat_blending_weights(molar_mass_ratio, oxygen2carbon)

    molar_mass_ratio = 0.5
    oxygen2carbon = -1.0
    with pytest.raises(ValueError):
        bat_blending_weights(molar_mass_ratio, oxygen2carbon)



"""
Particle Vapor Equilibrium, condensation and evaporation
based on partial pressures to get dm/dt or other forms of
particle growth and decay.

From Seinfeld and Pandas: The Condensation (chapter 13) Equation 13.3
This function calculates the rate of change of the mass of an aerosol
particle with diameter Dp.

The rate of change of mass, dm, is given by the formula:
    dm/dt = 4π * radius * Di * Mi * f(Kn, alpha) * delta_pi / RT
where:
    radius is the radius of the particle,
    Di is the diffusion coefficient of species i,
    Mi is the molar mass of species i,
    f(Kn, alpha) is a transition function of the Knudsen number (Kn) and the
    mass accommodation coefficient (alpha),
    delta_pi is the partial pressure of species i in the gas phase vs the
    particle phase.
    R is the gas constant,
    T is the temperature.

    An additional denominator term is added to acount for temperature changes,
    which is need for cloud droplets, but can be used in general too.

This is also in Aerosol Modeling Chapter 2 Equation 2.40

Seinfeld, J. H., & Pandis, S. N. (2016). Atmospheric Chemistry and Physics:
From Air Pollution to Climate Change. In Wiley (3rd ed.).
John Wiley & Sons, Inc.

Topping, D., & Bane, M. (2022). Introduction to Aerosol Modelling
(D. Topping & M. Bane, Eds.). Wiley. https://doi.org/10.1002/9781119625728

Units are all Base SI units.
"""

from typing import Union
from numpy.typing import NDArray
import numpy as np

# particula imports
from particula.util.constants import GAS_CONSTANT  # type: ignore


def first_order_mass_transport_k(
    radius: Union[float, NDArray[np.float64]],
    vapor_transition: Union[float, NDArray[np.float64]],
    diffusion_coefficient: Union[float, NDArray[np.float64]] = 2e-5,
) -> Union[float, NDArray[np.float64]]:
    """First-order mass transport coefficient per particle.

    Calculate the first-order mass transport coefficient, K, for a given radius
    diffusion coefficient, and vapor transition correction factor. For a
    single particle.

    Args:
        - radius : The radius of the particle [m].
        - diffusion_coefficient : The diffusion coefficient of the vapor
            [m^2/s], default to air.
        - vapor_transition : The vapor transition correction factor. [unitless]

    Returns:
        The first-order mass transport coefficient per particle (m^3/s).

    Examples:
        ``` py title="Float input"
        first_order_mass_transport_k(
            radius=1e-6,
            vapor_transition=0.6,
            diffusion_coefficient=2e-9
            )
        # Output: 1.5079644737231005e-14
        ```

        ``` py title="Array input"
        first_order_mass_transport_k
            radius=np.array([1e-6, 2e-6]),
            vapor_transition=np.array([0.6, 0.6]),
            diffusion_coefficient=2e-9
            )
        # Output: array([1.50796447e-14, 6.03185789e-14])
        ```
    References:
        - Aerosol Modeling: Chapter 2, Equation 2.49 (excluding number)
        - Mass Diffusivity:
            [Wikipedia](https://en.wikipedia.org/wiki/Mass_diffusivity)
    """
    if (
        isinstance(vapor_transition, np.ndarray)
        and vapor_transition.dtype == np.float64
        and vapor_transition.ndim == 2
    ):  # extent radius
        radius = radius[:, np.newaxis]  # type: ignore
    return (
        4 * np.pi * radius * diffusion_coefficient * vapor_transition
    )  # type: ignore


def mass_transfer_rate(
    pressure_delta: Union[float, NDArray[np.float64]],
    first_order_mass_transport: Union[float, NDArray[np.float64]],
    temperature: Union[float, NDArray[np.float64]],
    molar_mass: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """Calculate the mass transfer rate for a particle.

    Calculate the mass transfer rate based on the difference in partial
    pressure and the first-order mass transport coefficient.

    Args:
        - pressure_delta : The difference in partial pressure between the gas
            phase and the particle phase.
        - first_order_mass_transport : The first-order mass transport
            coefficient per particle.
        - temperature : The temperature at which the mass transfer rate is to
            be calculated.
        - molar_mass : The molar mass of the species [kg/mol].

    Returns:
        The mass transfer rate for the particle [kg/s].

    Examples:
        ``` py title="Single value input"
        mass_transfer_rate(
            pressure_delta=10.0,
            first_order_mass_transport=1e-17,
            temperature=300.0,
            molar_mass=0.02897
        )
        # Output: 1.16143004e-21
        ```

        ``` py title="Array input"
        mass_transfer_rate(
            pressure_delta=np.array([10.0, 15.0]),
            first_order_mass_transport=np.array([1e-17, 2e-17]),
            temperature=300.0,
            molar_mass=0.02897
        )
        # Output: array([1.16143004e-21, 3.48429013e-21])
        ```
    References:
        - Aerosol Modeling Chapter 2, Equation 2.41 (excluding particle number)
        - Seinfeld and Pandis: "Atmospheric Chemistry and Physics",
            Equation 13.3
    """
    return np.array(
        first_order_mass_transport
        * pressure_delta
        * molar_mass
        / (GAS_CONSTANT * temperature),
        dtype=np.float64,
    )


def radius_transfer_rate(
    mass_rate: Union[float, NDArray[np.float64]],
    radius: Union[float, NDArray[np.float64]],
    density: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """Convert mass rate to radius transfer rate.

    Convert the mass rate to a radius transfer rate based on the
    volume of the particle.

    Args:
        - mass_rate : The mass transfer rate for the particle [kg/s].
        - radius : The radius of the particle [m].
        - density : The density of the particle [kg/m^3].

    Returns:
        The radius growth rate for the particle [m/s].

    Examples:
        ``` py title="Single value input"
        radius_transfer_rate(
            mass_rate=1e-21,
            radius=1e-6,
            density=1000
        )
        # Output: 7.95774715e-14
        ```

        ``` py title="Array input"
        radius_transfer_rate(
            mass_rate=np.array([1e-21, 2e-21]),
            radius=np.array([1e-6, 2e-6]),
            density=1000
        )
        # Output: array([7.95774715e-14, 1.98943679e-14])
        ```
    """
    if isinstance(mass_rate, np.ndarray) and mass_rate.ndim == 2:
        radius = radius[:, np.newaxis]  # type: ignore
    return mass_rate / (density * 4 * np.pi * radius**2)  # type: ignore


def calculate_mass_transfer(
    mass_rate: NDArray[np.float64],
    time_step: float,
    gas_mass: NDArray[np.float64],
    particle_mass: NDArray[np.float64],
    particle_concentration: NDArray[np.float64],
) -> NDArray[np.float64]:
    """
    Helper function that routes the mass transfer calculation to either the
    single-species or multi-species calculation functions based on the input
    dimensions of gas_mass.

    Args:
        - mass_rate : The rate of mass transfer per particle (kg/s).
        - time_step : The time step for the mass transfer calculation (sec).
        - gas_mass : The available mass of gas species (kg).
        - particle_mass : The mass of each particle (kg).
        - particle_concentration : The concentration of particles (number/m^3).

    Returns:
        The amount of mass transferred, accounting for gas and particle
            limitations.

    Examples:
        ``` py title="Single species input"
        calculate_mass_transfer(
            mass_rate=np.array([0.1, 0.5]),
            time_step=10,
            gas_mass=np.array([0.5]),
            particle_mass=np.array([1.0, 50]),
            particle_concentration=np.array([1, 0.5])
        )
        ```

        ``` py title="Multiple species input"
        calculate_mass_transfer(
            mass_rate=np.array([[0.1, 0.05, 0.03], [0.2, 0.15, 0.07]]),
            time_step=10,
            gas_mass=np.array([1.0, 0.8, 0.5]),
            particle_mass=np.array([[1.0, 0.9, 0.8], [1.2, 1.0, 0.7]]),
            particle_concentration=np.array([5, 4])
        )
        ```
    """
    if gas_mass.size == 1:  # Single species case
        return calculate_mass_transfer_single_species(
            mass_rate,
            time_step,
            gas_mass,
            particle_mass,
            particle_concentration,
        )
    # Multiple species case
    return calculate_mass_transfer_multiple_species(
        mass_rate,
        time_step,
        gas_mass,
        particle_mass,
        particle_concentration,
    )


def calculate_mass_transfer_single_species(
    mass_rate: NDArray[np.float64],
    time_step: float,
    gas_mass: NDArray[np.float64],
    particle_mass: NDArray[np.float64],
    particle_concentration: NDArray[np.float64],
) -> NDArray[np.float64]:
    """
    Calculate mass transfer for a single gas species (m=1).

    Args:
        - mass_rate : The rate of mass transfer per particle (number*kg/s).
        - time_step : The time step for the mass transfer calculation (sec).
        - gas_mass : The available mass of gas species (kg).
        - particle_mass : The mass of each particle (kg).
        - particle_concentration : The concentration of particles (number/m^3).

    Returns:
        The amount of mass transferred for a single gas species.

    Examples:
        ``` py title="Single species input"
        calculate_mass_transfer_single_species(
            mass_rate=np.array([0.1, 0.5]),
            time_step=10,
            gas_mass=np.array([0.5]),
            particle_mass=np.array([1.0, 50]),
            particle_concentration=np.array([1, 0.5])
        )
        ```
    """
    # Step 1: Calculate the total mass to be transferred
    # (accounting for particle concentration)
    mass_to_change = mass_rate * time_step * particle_concentration
    # Step 2: Calculate the total requested mass
    total_requested_mass = mass_to_change.sum()
    # Step 3: Scale the mass if total requested mass exceeds available gas mass
    if total_requested_mass > gas_mass:
        scaling_factor = gas_mass / total_requested_mass
        mass_to_change *= scaling_factor
    # Step 4: Limit condensation by available gas mass
    condensible_mass_transfer = np.minimum(mass_to_change, gas_mass)
    # Step 5: Limit evaporation by available particle mass
    evaporative_mass_transfer = np.maximum(
        mass_to_change, -particle_mass * particle_concentration
    )
    # Step 6: Determine final transferable mass (condensation or evaporation)
    transferable_mass = np.where(
        mass_to_change > 0,  # Condensation scenario
        condensible_mass_transfer,  # Limited by gas mass
        evaporative_mass_transfer,  # Limited by particle mass
    )
    return transferable_mass


def calculate_mass_transfer_multiple_species(
    mass_rate: NDArray[np.float64],
    time_step: float,
    gas_mass: NDArray[np.float64],
    particle_mass: NDArray[np.float64],
    particle_concentration: NDArray[np.float64],
) -> NDArray[np.float64]:
    """
    Calculate mass transfer for multiple gas species.

    Args:
        - mass_rate : The rate of mass transfer per particle for each gas
            species (kg/s).
        - time_step : The time step for the mass transfer calculation (sec).
        - gas_mass : The available mass of each gas species (kg).
        - particle_mass : The mass of each particle for each gas species (kg).
        - particle_concentration : The concentration of particles for each gas
            species (number/m^3).

    Returns:
        The amount of mass transferred for multiple gas species.

    Examples:
        ``` py title="Multiple species input"
        calculate_mass_transfer_multiple_species(
            mass_rate=np.array([[0.1, 0.05, 0.03], [0.2, 0.15, 0.07]]),
            time_step=10,
            gas_mass=np.array([1.0, 0.8, 0.5]),
            particle_mass=np.array([[1.0, 0.9, 0.8], [1.2, 1.0, 0.7]]),
            particle_concentration=np.array([5, 4])
        )
        ```
    """
    # Step 1: Calculate the total mass to change
    # (considering particle concentration)
    mass_to_change = (
        mass_rate * time_step * particle_concentration[:, np.newaxis]
    )

    # Step 2: Total requested mass for each gas species (sum over particles)
    total_requested_mass = mass_to_change.sum(axis=0)

    # Step 3: Create scaling factors where requested mass exceeds available
    # gas mass
    scaling_factors = np.ones_like(mass_to_change)
    scaling_mask = total_requested_mass > gas_mass

    # Apply scaling where needed (scaling along the gas species axis)
    scaling_factors[:, scaling_mask] = (
        gas_mass[scaling_mask] / total_requested_mass[scaling_mask]
    )

    # Step 4: Apply scaling factors to the mass_to_change
    mass_to_change *= scaling_factors

    # Step 5: Limit condensation by available gas mass
    condensible_mass_transfer = np.minimum(np.abs(mass_to_change), gas_mass)

    # Step 6: Limit evaporation by available particle mass
    evaporative_mass_transfer = np.maximum(
        mass_to_change, -particle_mass * particle_concentration[:, np.newaxis]
    )

    # Step 7: Determine the final transferable mass
    # (condensation or evaporation)
    transferable_mass = np.where(
        mass_to_change > 0,  # Condensation scenario
        condensible_mass_transfer,  # Limited by gas mass
        evaporative_mass_transfer,  # Limited by particle mass
    )

    return transferable_mass



"""Module for the vapor transition correction function, which accounts for the
intermediate regime between continuum and free molecular flow. This is the
Suchs and Futugin transition function.
"""

from typing import Union
from numpy.typing import NDArray
import numpy as np


def vapor_transition_correction(
    knudsen_number: Union[float, NDArray[np.float64]],
    mass_accommodation: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """
    Calculate the transition correction factor, f(Kn, alpha), for a given
    Knudsen number and mass accommodation coefficient. This function is used to
    account for the intermediate regime between continuum and free molecular
    flow. This is the Suchs and Futugin transition function.

    Args:
    -----
    - knudsen_number (Union[float, NDArray[np.float64]]): The Knudsen number,
    which quantifies the relative importance of the mean free path of gas
    molecules to the size of the particle.
    - mass_accommodation (Union[float, NDArray[np.float64]]): The mass
    accommodation coefficient, representing the probability of a gas molecule
    sticking to the particle upon collision.

    Returns:
    --------
    - Union[float, NDArray[np.float64]]: The transition correction value
    calculated based on the specified inputs.

    References:
    ----------
    - Seinfeld and Pandis, "Atmospheric Chemistry and Physics", Chapter 12,
    equation 12.43.
    Note: There are various formulations for this correction, so further
    extensions of this function might be necessary depending on specific
    requirements.
    - Original reference:
    FUCHS, N. A., & SUTUGIN, A. G. (1971). HIGH-DISPERSED AEROSOLS.
    In Topics in Current Aerosol Research (p. 1). Elsevier.
    https://doi.org/10.1016/B978-0-08-016674-2.50006-6
    """
    return (0.75 * mass_accommodation * (1 + knudsen_number)) / (
        (knudsen_number**2 + knudsen_number)
        + 0.283 * mass_accommodation * knudsen_number
        + 0.75 * mass_accommodation
    )



"""
Calculate the particle Reynolds number.
"""

from typing import Union
from numpy.typing import NDArray
import numpy as np

from particula.util.validate_inputs import validate_inputs


@validate_inputs(
    {
        "particle_radius": "positive",
        "particle_velocity": "positive",
        "kinematic_viscosity": "positive",
    }
)
def get_particle_reynolds_number(
    particle_radius: Union[float, NDArray[np.float64]],
    particle_velocity: Union[float, NDArray[np.float64]],
    kinematic_viscosity: Union[float, NDArray[np.float64]],
) -> Union[float, NDArray[np.float64]]:
    """
    Calculate the particle Reynolds number.

    The particle Reynolds number (Re_p) characterizes the flow behavior around
    a particle in a fluid and is given by:

        Re_p = (2 a v_p) / ν

    - Re_p : Particle Reynolds number [-]
    - a (particle_radius) : Particle radius [m]
    - v_p (particle_velocity) : Particle velocity relative to the fluid
        [m/s]
    - ν (kinematic_viscosity) : Kinematic viscosity of the fluid [m²/s]



    Arguments:
    ----------
        - particle_radius : Radius of the particle [m]
        - particle_velocity : Particle velocity relative to the fluid [m/s]
        - kinematic_viscosity : Kinematic viscosity of the surrounding fluid
            [m²/s]

    Returns:
    --------
        - Particle Reynolds number [-]

    References:
    -----------
        - Wikipedia https://en.wikipedia.org/wiki/Reynolds_number
        - **Stokes Flow (Viscous Dominated, Re_p < 1)**:
            - Particles follow the fluid closely (e.g., aerosols).
        - **Transitional Flow (1 < Re_p < 1000)**:
            - Both **viscous and inertial forces** contribute to flow behavior.
            - Intermediate drag corrections apply.
        - **Turbulent Flow (Re_p > 1000)**:
            - **Inertial forces dominate**, resulting in vortex shedding and
                wake formation.
            - Applies to **large, fast-moving particles**
                (e.g., raindrops, large sediment).
    """
    return (2 * particle_radius * particle_velocity) / kinematic_viscosity



"""Test for transition_regime calculation functions."""

import numpy as np
import pytest

from particula.dynamics.coagulation import transition_regime

# test single value
DIFFUSIVE_KNUDSEN_SINGLE = 0.1
COULOMB_POTENTIAL_RATIO_SINGLE = 0.5

# test array
DIFFUSIVE_KNUDSEN_ARRAY = np.array([0.5, 1.0, 5.0, 10.0])
COULOMB_POTENTIAL_RATIO_ARRAY = np.array([0.7, 0.9, 1.0, 1.1])


def test_hard_sphere():
    """
    Test the hard_sphere function with a single value and
    array of diffusive_knudsen values.
    """
    # single value
    expected_single = 0.10960430161885967
    result_single = transition_regime.hard_sphere(DIFFUSIVE_KNUDSEN_SINGLE)
    np.testing.assert_almost_equal(result_single, expected_single, decimal=4)
    # array
    expected = np.array([1.65971644, 4.12694075, 24.16690909, 49.22484307])
    result = transition_regime.hard_sphere(DIFFUSIVE_KNUDSEN_ARRAY)
    np.testing.assert_almost_equal(result, expected, decimal=4)


def test_coulomb_dyachkov2007():
    """
    Test the coulomb_dyachkov2007 function with a single value and
    array of diffusive_knudsen and coulomb_potential_ratio values.
    """
    # single value
    expected_single = 0.10960430161885967
    result_single = transition_regime.coulomb_dyachkov2007(
        DIFFUSIVE_KNUDSEN_SINGLE, COULOMB_POTENTIAL_RATIO_SINGLE
    )
    np.testing.assert_almost_equal(result_single, expected_single, decimal=4)
    # array
    expected = np.array([1.73703563, 4.60921277, 26.22159795, 51.92102133])
    result = transition_regime.coulomb_dyachkov2007(
        DIFFUSIVE_KNUDSEN_ARRAY, COULOMB_POTENTIAL_RATIO_ARRAY
    )
    np.testing.assert_almost_equal(result, expected, decimal=4)


def test_coulomb_gatti2008():
    """
    Test the coulomb_gatti2008 function with a single value and
    array of diffusive_knudsen and coulomb_potential_ratio values.
    """
    # single value
    expected_single = 0.12621027
    result_single = transition_regime.coulomb_gatti2008(
        DIFFUSIVE_KNUDSEN_SINGLE, COULOMB_POTENTIAL_RATIO_SINGLE
    )
    np.testing.assert_almost_equal(result_single, expected_single, decimal=4)
    # array
    expected = np.array([2.00132915, 5.10865767, 26.42422258, 52.43789491])
    result = transition_regime.coulomb_gatti2008(
        DIFFUSIVE_KNUDSEN_ARRAY, COULOMB_POTENTIAL_RATIO_ARRAY
    )
    np.testing.assert_almost_equal(result, expected, decimal=4)


def test_coulomb_gopalakrishnan2012():
    """
    Test the coulomb_gopalakrishnan2012 function with a single value and
    array of diffusive_knudsen and coulomb_potential_ratio values.
    """
    # single value
    expected_single = 0.1096043
    result_single = transition_regime.coulomb_gopalakrishnan2012(
        DIFFUSIVE_KNUDSEN_SINGLE, COULOMB_POTENTIAL_RATIO_SINGLE
    )
    np.testing.assert_almost_equal(result_single, expected_single, decimal=4)
    # array
    expected = np.array([1.83746548, 4.83694019, 24.16690909, 49.22484307])
    result = transition_regime.coulomb_gopalakrishnan2012(
        DIFFUSIVE_KNUDSEN_ARRAY, COULOMB_POTENTIAL_RATIO_ARRAY
    )
    np.testing.assert_almost_equal(result, expected, decimal=4)


def test_coulomb_chahl2019():
    """
    Test the coulomb_chahl2019 function with a single value and
    array of diffusive_knudsen and coulomb_potential_ratio values.
    """
    # single value
    expected_single = 0.10960430161885967
    result_single = transition_regime.coulomb_chahl2019(
        DIFFUSIVE_KNUDSEN_SINGLE, COULOMB_POTENTIAL_RATIO_SINGLE
    )
    np.testing.assert_almost_equal(result_single, expected_single, decimal=4)
    # array
    expected = np.array([1.65863442, 4.37444613, 28.05501739, 59.74082667])
    result = transition_regime.coulomb_chahl2019(
        DIFFUSIVE_KNUDSEN_ARRAY, COULOMB_POTENTIAL_RATIO_ARRAY
    )
    np.testing.assert_almost_equal(result, expected, decimal=4)


def test_transition_regime_coagulation_edge_cases():
    """
    Test the transition_regime_coagulation function with edge case values.

    Edge cases tested:
    - Zero values: Should handle zero particle sizes correctly
    - Small values (0.1): Should compute valid coagulation rates
    - Negative values: Should raise ValueError (physically impossible)
    - Very large values: Should compute without numerical overflow
    - NaN: Should raise ValueError (invalid input)
    - Infinity: Should raise ValueError (invalid input)
    """
    # Test zero and small values
    small_zero_input = np.array([0.1, 0.0])
    small_zero_result = transition_regime.hard_sphere(small_zero_input)
    small_zero_expected = np.array([0.10960430161885967, 0.0])
    np.testing.assert_almost_equal(small_zero_result, small_zero_expected, decimal=4)

    # Test very large values (using realistic upper bound for particle sizes)
    large_input = np.array([1e5, 1e5])  # 100mm particles
    large_result = transition_regime.hard_sphere(large_input)
    assert np.isfinite(large_result).all(), "Should handle large values without overflow"

    # Test negative values
    negative_input = np.array([-1.0, 1.0])
    with pytest.raises(ValueError, match="Particle sizes must be non-negative"):
        transition_regime.hard_sphere(negative_input)

    # Test NaN values
    nan_input = np.array([np.nan, 1.0])
    with pytest.raises(ValueError, match="Invalid particle sizes"):
        transition_regime.hard_sphere(nan_input)

    # Test infinity values
    inf_input = np.array([np.inf, 1.0])
    with pytest.raises(ValueError, match="Invalid particle sizes"):
        transition_regime.hard_sphere(inf_input)



"""Builds distributions strategies based on the specified representation.

Currently, there are no parameters to set, but this is used for consistency
with other builder patterns in the codebase.
"""

from particula.particles.distribution_strategies import (
    MassBasedMovingBin,
    RadiiBasedMovingBin,
    SpeciatedMassMovingBin,
    ParticleResolvedSpeciatedMass,
)
from particula.abc_builder import BuilderABC


class MassBasedMovingBinBuilder(BuilderABC):
    """Builds a MassBasedMovingBin instance."""

    def __init__(self) -> None:
        required_parameters = None
        BuilderABC.__init__(self, required_parameters)

    def build(self) -> MassBasedMovingBin:
        """Builds a MassBasedMovingBin instance."""
        return MassBasedMovingBin()


class RadiiBasedMovingBinBuilder(BuilderABC):
    """Builds a RadiiBasedMovingBin instance."""

    def __init__(self) -> None:
        required_parameters = None
        BuilderABC.__init__(self, required_parameters)

    def build(self) -> RadiiBasedMovingBin:
        """Builds a RadiiBasedMovingBin instance."""
        return RadiiBasedMovingBin()


class SpeciatedMassMovingBinBuilder(BuilderABC):
    """Builds a SpeciatedMassMovingBin instance."""

    def __init__(self) -> None:
        required_parameters = None
        BuilderABC.__init__(self, required_parameters)

    def build(self) -> SpeciatedMassMovingBin:
        """Builds a SpeciatedMassMovingBin instance."""
        return SpeciatedMassMovingBin()


class ParticleResolvedSpeciatedMassBuilder(BuilderABC):
    """Builds a ParticleResolvedSpeciatedMass instance."""

    def __init__(self) -> None:
        required_parameters = None
        BuilderABC.__init__(self, required_parameters)

    def build(self) -> ParticleResolvedSpeciatedMass:
        return ParticleResolvedSpeciatedMass()



"""Tests for the mean_thermal_speed function."""

import pytest
import numpy as np
from particula.particles.properties import mean_thermal_speed
from particula.util.constants import BOLTZMANN_CONSTANT


def test_mean_thermal_speed_single_value():
    """
    Test mean_thermal_speed with single float inputs for mass and temperature.
    """
    mass = 4.188790204786391e-12  # Example mass in kg
    temperature = 300  # Temperature in Kelvin
    expected_speed = np.sqrt(
        (8 * BOLTZMANN_CONSTANT * temperature) / (np.pi * mass)
    )
    assert np.isclose(mean_thermal_speed(mass, temperature), expected_speed)


def test_mean_thermal_speed_array_input():
    """
    Test mean_thermal_speed with numpy array inputs for mass and temperature.
    """
    mass = np.array([2.18e-25, 2.18e-25])
    temperature = np.array([300, 350])
    expected_speed = np.sqrt(
        (8 * BOLTZMANN_CONSTANT * temperature) / (np.pi * mass)
    )
    assert np.allclose(
        mean_thermal_speed(mass, temperature), expected_speed, rtol=1e-6
    )


def test_mean_thermal_speed_input_validation():
    """
    Ensure that providing incorrect input types to mean_thermal_speed raises
    a TypeError.
    """
    with pytest.raises(TypeError):
        mean_thermal_speed("not a number", "also not a number")



"""
Kinematic viscosity for fluids.

The kinematic viscosity is the ratio of the dynamic viscosity to the
density of a fluid. It is a measure of the fluid's resistance to flow
under the influence of gravity.

v = mu / rho

Where:
    - v : Kinematic viscosity [m^2/s].
    - mu : Dynamic viscosity [Pa*s].
    - rho : Density of the fluid [kg/m^3].

References:
https://resources.wolframcloud.com/FormulaRepository/resources/Viscosity-Conversion-Formula
https://en.wikipedia.org/wiki/Viscosity#Kinematic_viscosity
"""

from particula.util.validate_inputs import validate_inputs
from particula.gas.properties.dynamic_viscosity import get_dynamic_viscosity
from particula.util.constants import (REF_TEMPERATURE_STP,
                                      REF_VISCOSITY_AIR_STP,)


def get_kinematic_viscosity(
    dynamic_viscosity: float,
    fluid_density: float,
) -> float:
    """
    Calculate the kinematic viscosity of a fluid.

    The kinematic viscosity is the ratio of the dynamic viscosity to the
    density of a fluid. It is a measure of the fluid's resistance to flow
    under the influence of gravity.

    Args:
    -----
        - dynamic_viscosity : Dynamic viscosity of the fluid [Pa*s].
        - fluid_density : Density of the fluid [kg/m³].

    Returns:
    --------
        - Kinematic viscosity of the fluid [m^2/s].

    References:
    ----------
    - Equation : v = mu / rho
            - v : Kinematic viscosity [m^2/s].
            - mu : Dynamic viscosity [Pa*s].
            - rho : Density of the fluid [kg/m^3].
    https://resources.wolframcloud.com/FormulaRepository/resources/Viscosity-Conversion-Formula
    """
    return dynamic_viscosity / fluid_density


@validate_inputs({"temperature": "positive"})
def get_kinematic_viscosity_via_system_state(
    temperature: float,
    fluid_density: float,
    reference_viscosity: float = REF_VISCOSITY_AIR_STP,
    reference_temperature: float = REF_TEMPERATURE_STP
) -> float:
    """
    Calculate the kinematic viscosity of air by calculating dynamic viscosity
    of air.

    Args:
    -----
    - temperature: Desired air temperature [K]. Must be greater than 0.
    - fluid_density: Density of the fluid [kg/m³].
    - reference_viscosity: Gas viscosity [Pa*s] at the reference temperature
    (default is STP).
    - reference_temperature: Gas temperature [K] for the reference viscosity
    (default is STP).

    Returns:
    --------
    - float: The kinematic viscosity of air at the given temperature [m^2/s].

    Raises:
    ------
    - ValueError: If the temperature is less than or equal to 0.

    References:
    ----------
    https://resources.wolframcloud.com/FormulaRepository/resources/Sutherlands-Formula
    """
    dynamic_viscosity = get_dynamic_viscosity(
        temperature=temperature,
        reference_viscosity=reference_viscosity,
        reference_temperature=reference_temperature
    )
    return get_kinematic_viscosity(
        dynamic_viscosity=dynamic_viscosity,
        fluid_density=fluid_density
        )



"""
Calculates the Gibbs free energy of mixing for a binary solution.
"""

from typing import Tuple

from numpy.typing import NDArray
import numpy as np

from particula.util.machine_limit import safe_log


def gibbs_free_engery(
    organic_mole_fraction: NDArray[np.float64],
    gibbs_mix: NDArray[np.float64],
) -> Tuple[
    NDArray[np.float64],
    NDArray[np.float64]
]:
    """
    Calculate the gibbs free energy of the mixture. Ideal and non-ideal.

    Args:
        - organic_mole_fraction : A numpy array of organic mole
            fractions.
        - gibbs_mix : A numpy array of gibbs free energy of mixing.

    Returns:
        - gibbs_ideal : The ideal gibbs free energy of mixing.
        - gibbs_real : The real gibbs free energy of mixing.
    """

    gibbs_ideal = (1 - organic_mole_fraction) \
        * safe_log(1 - organic_mole_fraction) \
        + organic_mole_fraction \
        * safe_log(organic_mole_fraction)
    gibbs_real = gibbs_ideal + gibbs_mix
    return gibbs_ideal, gibbs_real



"""Particle representation."""

from typing import Optional
from copy import deepcopy
import logging
import numpy as np
from numpy.typing import NDArray

# From Particula
from particula.particles.activity_strategies import ActivityStrategy
from particula.particles.surface_strategies import SurfaceStrategy
from particula.particles.distribution_strategies import (
    DistributionStrategy,
)

logger = logging.getLogger("particula")


# pylint: disable=too-many-instance-attributes
class ParticleRepresentation:
    """Everything needed to represent a particle or a collection of particles.

    Represents a particle or a collection of particles, encapsulating the
    strategy for calculating mass, radius, and total mass based on a
    specified particle distribution, density, and concentration. This class
    allows for flexibility in representing particles.

    Attributes:
        - strategy : The computation strategy for particle representations.
        - activity : The activity strategy for the partial pressure
            calculations.
        - surface : The surface strategy for surface tension and Kelvin effect.
        - distribution : The distribution data for the particles, which could
            represent sizes, masses, or another relevant metric.
        - density : The density of the material from which the particles are
            made.
        - concentration : The concentration of particles within the
            distribution.
        - charge : The charge on each particle.
        - volume : The air volume for simulation of particles in the air,
            default is 1 m^3. This is only used in ParticleResolved Strategies.
    """

    def __init__(
        self,
        strategy: DistributionStrategy,
        activity: ActivityStrategy,
        surface: SurfaceStrategy,
        distribution: NDArray[np.float64],
        density: NDArray[np.float64],
        concentration: NDArray[np.float64],
        charge: NDArray[np.float64],
        volume: float = 1,
    ):  # pylint: disable=too-many-positional-arguments, too-many-arguments
        self.strategy = strategy
        self.activity = activity
        self.surface = surface
        self.distribution = distribution
        self.density = density
        self.concentration = concentration
        self.charge = charge
        self.volume = volume

    def __str__(self) -> str:
        """Returns a string representation of the particle representation.

        Returns:
            - str : A string representation of the particle representation.

        Example:
            ``` py title="Get String Representation"
            str_rep = str(particle_representation)
            print(str_rep)
            ```
        """
        return (
            f"Particle Representation:\n"
            f"\tStrategy: {self.get_strategy_name()}\n"
            f"\tActivity: {self.get_activity_name()}\n"
            f"\tSurface: {self.get_surface_name()}\n"
            f"\tMass Concentration: "
            f"{self.get_mass_concentration():.3e} [kg/m^3]\n"
            f"\tNumber Concentration: "
            f"{self.get_total_concentration():.3e} [#/m^3]"
        )

    def get_strategy(self, clone: bool = False) -> DistributionStrategy:
        """Returns the strategy used for particle representation.

        Args:
            - clone : If True, then return a deepcopy of the strategy.

        Returns:
            - The strategy used for particle representation.

        Example:
            ``` py title="Get Strategy"
            strategy = particle_representation.get_strategy()
            ```
        """
        if clone:
            return deepcopy(self.strategy)
        return self.strategy

    def get_strategy_name(self) -> str:
        """Returns the name of the strategy used for particle representation.

        Returns:
            - The name of the strategy used for particle representation.

        Example:
            ``` py title="Get Strategy Name"
            strategy_name = particle_representation.get_strategy_name()
            print(strategy_name)
            ```
        """
        return self.strategy.get_name()

    def get_activity(self, clone: bool = False) -> ActivityStrategy:
        """Returns the activity strategy used for partial pressure
        calculations.

        Args:
            - clone : If True, then return a deepcopy of the activity strategy.

        Returns:
            - The activity strategy used for partial pressure calculations.

        Example:
            ``` py title="Get Activity Strategy"
            activity = particle_representation.get_activity()
            ```
        """
        if clone:
            return deepcopy(self.activity)
        return self.activity

    def get_activity_name(self) -> str:
        """Returns the name of the activity strategy used for partial pressure
        calculations.

        Returns:
            - The name of the activity strategy used for partial pressure
              calculations.

        Example:
            ``` py title="Get Activity Strategy Name"
            activity_name = particle_representation.get_activity_name()
            print(activity_name)
            ```
        """
        return self.activity.get_name()

    def get_surface(self, clone: bool = False) -> SurfaceStrategy:
        """Returns the surface strategy used for surface tension and
        Kelvin effect.

        Args:
            - clone : If True, then return a deepcopy of the surface strategy.

        Returns:
            - The surface strategy used for surface tension and Kelvin effect.

        Example:
            ``` py title="Get Surface Strategy"
            surface = particle_representation.get_surface()
            ```
        """
        if clone:
            return deepcopy(self.surface)
        return self.surface

    def get_surface_name(self) -> str:
        """Returns the name of the surface strategy used for surface tension
        and Kelvin effect.

        Returns:
            - The name of the surface strategy used for surface tension and
                Kelvin effect.

        Example:
            ``` py title="Get Surface Strategy Name"
            surface_name = particle_representation.get_surface_name()
            print(surface_name)
            ```
        """
        return self.surface.get_name()

    def get_distribution(self, clone: bool = False) -> NDArray[np.float64]:
        """Returns the distribution of the particles.

        Args:
            - clone : If True, then return a copy of the distribution array.

        Returns:
            - The distribution of the particles.

        Example:
            ``` py title="Get Distribution Array"
            distribution = particle_representation.get_distribution()
            ```
        """
        if clone:
            return np.copy(self.distribution)
        return self.distribution

    def get_density(self, clone: bool = False) -> NDArray[np.float64]:
        """Returns the density of the particles.

        Args:
            - clone : If True, then return a copy of the density array.

        Returns:
            - The density of the particles.

        Example:
            ``` py title="Get Density Array"
            density = particle_representation.get_density()
            ```
        """
        if clone:
            return np.copy(self.density)
        return self.density

    def get_concentration(self, clone: bool = False) -> NDArray[np.float64]:
        """Returns the volume concentration of the particles.

        For ParticleResolved Strategies, the concentration is the number of
        particles per self.volume to get concentration/m^3. For other
        Strategies, the concentration is the already per 1/m^3.

        Args:
            - clone : If True, then return a copy of the concentration array.

        Returns:
            - The concentration of the particles.

        Example:
            ``` py title="Get Concentration Array"
            concentration = particle_representation.get_concentration()
            ```
        """
        if clone:
            return np.copy(self.concentration / self.volume)
        return self.concentration / self.volume

    def get_total_concentration(self, clone: bool = False) -> np.float64:
        """Returns the total concentration of the particles.

        Args:
            - clone : If True, then return a copy of the concentration array.

        Returns:
            - The concentration of the particles.

        Example:
            ``` py title="Get Total Concentration"
            total_concentration = (
                particle_representation.get_total_concentration()
            )
            print(total_concentration)
            ```
        """
        return np.sum(self.get_concentration(clone=clone))

    def get_charge(self, clone: bool = False) -> NDArray[np.float64]:
        """Returns the charge per particle.

        Args:
            - clone : If True, then return a copy of the charge array.

        Returns:
            - The charge of the particles.

        Example:
            ``` py title="Get Charge Array"
            charge = particle_representation.get_charge()
            ```
        """
        if clone:
            return np.copy(self.charge)
        return self.charge

    def get_volume(self, clone: bool = False) -> float:
        """Returns the volume of the particles.

        Args:
            - clone : If True, then return a copy of the volume array.

        Returns:
            - The volume of the particles.

        Example:
            ``` py title="Get Volume"
            volume = particle_representation.get_volume()
            ```
        """
        if clone:
            return deepcopy(self.volume)
        return self.volume

    def get_species_mass(self, clone: bool = False) -> NDArray[np.float64]:
        """Returns the masses per species in the particles.

        Args:
            - clone : If True, then return a copy of the mass array.

        Returns:
            - The mass of the particles per species.

        Example:
            ``` py title="Get Species Mass"
            species_mass = particle_representation.get_species_mass()
            ```
        """
        if clone:
            return np.copy(
                self.strategy.get_species_mass(self.distribution, self.density)
            )
        return self.strategy.get_species_mass(self.distribution, self.density)

    def get_mass(self, clone: bool = False) -> NDArray[np.float64]:
        """Returns the mass of the particles as calculated by the strategy.

        Args:
            - clone : If True, then return a copy of the mass array.

        Returns:
            - The mass of the particles.

        Example:
            ``` py title="Get Mass"
            mass = particle_representation.get_mass()
            ```
        """
        if clone:
            return np.copy(
                self.strategy.get_mass(self.distribution, self.density)
            )
        return self.strategy.get_mass(self.distribution, self.density)

    def get_mass_concentration(self, clone: bool = False) -> np.float64:
        """Returns the total mass / volume simulated.

        The mass concentration is as calculated by the strategy, taking into
        account the distribution and concentration.

        Args:
            - clone : If True, then return a copy of the mass concentration.

        Returns:
            - np.float64 : The mass concentration of the particles, kg/m^3.

        Example:
            ``` py title="Get Mass Concentration"
            mass_concentration = (
                particle_representation.get_mass_concentration()
            )
            print(mass_concentration)
            ```
        """
        if clone:
            return deepcopy(
                self.strategy.get_total_mass(
                    self.get_distribution(),
                    self.get_concentration(),
                    self.get_density(),
                )
            )
        return self.strategy.get_total_mass(
            self.get_distribution(),
            self.get_concentration(),
            self.get_density(),
        )

    def get_radius(self, clone: bool = False) -> NDArray[np.float64]:
        """Returns the radius of the particles as calculated by the strategy.

        Args:
            - clone : If True, then return a copy of the radius array.

        Returns:
            - The radius of the particles.

        Example:
            ``` py title="Get Radius"
            radius = particle_representation.get_radius()
            ```
        """
        if clone:
            return np.copy(
                self.strategy.get_radius(self.distribution, self.density)
            )
        return self.strategy.get_radius(self.distribution, self.density)

    def add_mass(self, added_mass: NDArray[np.float64]) -> None:
        """Adds mass to the particle distribution, and updates parameters.

        Args:
            - added_mass : The mass to be added per
              distribution bin.

        Example:
            ``` py title="Add Mass"
            particle_representation.add_mass(added_mass)
            ```
        """
        # maybe remove return concentration
        (self.distribution, _) = self.strategy.add_mass(
            self.get_distribution(),
            self.get_concentration(),
            self.get_density(),
            added_mass,
        )

    def add_concentration(
        self,
        added_concentration: NDArray[np.float64],
        added_distribution: Optional[NDArray[np.float64]] = None,
    ) -> None:
        """Adds concentration to the particle distribution.

        Args:
            - added_concentration : The concentration to be
              added per distribution bin.

        Example:
            ``` py title="Add Concentration"
            particle_representation.add_concentration(added_concentration)
            ```
        """
        # if added_distribution is None, then it will be calculated
        if added_distribution is None:
            message = "Added distribution is value None."
            logger.warning(message)
            added_distribution = self.get_distribution()
        # self.concentration += added_concentration
        (self.distribution, self.concentration) = (
            self.strategy.add_concentration(
                distribution=self.get_distribution(),
                concentration=self.get_concentration(),
                added_distribution=added_distribution,
                added_concentration=added_concentration,
            )
        )

    def collide_pairs(self, indices: NDArray[np.int64]) -> None:
        """Collide pairs of indices, used for ParticleResolved Strategies.

        Args:
            - indices : The indices to collide.

        Example:
            ``` py title="Collide Pairs"
            particle_representation.collide_pairs(indices)
            ```
        """
        (self.distribution, self.concentration) = self.strategy.collide_pairs(
            self.distribution, self.concentration, self.density, indices
        )



"""Radius of collision module for non-spherical particles. Based on the
compiled models in Qian et al. (2022).

Qian, W., Kronenburg, A., Hui, X., Lin, Y., & Karsch, M. (2022).
Effects of agglomerate characteristics on their collision kernels in
the free molecular regime. Journal of Aerosol Science, 159.
https://doi.org/10.1016/j.jaerosci.2021.10586
"""

from typing import Union
import numpy as np
from numpy.typing import NDArray


def mulholland_1988(
    radius_gyration: Union[NDArray[np.float64], float]
) -> Union[NDArray[np.float64], float]:
    """Collision radius is equal to the radius of gyration.

    Args:
        radius_gyration: Radius of gyration of the particle [m].

    Returns:
        (float or NDArray[float]): Collision radius of the particle [m].

    References:
        Mulholland, G. W., Mountain, R. D., Samson, R. J., & Ernst, M. H.
        (1988). Cluster Size Distribution for Free Molecular Agglomeration.
        Energy and Fuels, 2(4). https://doi.org/10.1021/ef00010a014

    Examples:
        ``` py title="Example"
        mulholland_1988(1.5)
        # 1.5
        ```
    """
    return radius_gyration


def rogak_flagan_1992(
    radius_gyration: Union[NDArray[np.float64], float],
    fractal_dimension: Union[NDArray[np.float64], float],
) -> Union[NDArray[np.float64], float]:
    """Collision radius with fractal dimension by Rogak and Flagan 1992.

    Args:
        radius_gyration: Radius of gyration of the particle [m].
        fractal_dimension: Fractal dimension of the particle
            [dimensionless, df].

    Returns:
        (float or NDArray[float]): Collision radius of the particle [m].

    References:
        Rogak, S. N., & Flagan, R. C. (1992). Coagulation of aerosol
        agglomerates in the transition regime. Journal of Colloid and
        Interface Science, 151(1), 203-224.
        https://doi.org/10.1016/0021-9797(92)90252-H

    Examples:
        ``` py title="Example"
        rogak_flagan_1992(1.5, 2.5)
        # 1.83711
        ```
    """
    return np.sqrt((fractal_dimension + 2) / 3) * radius_gyration


def zurita_gotor_2002(
    radius_gyration: Union[NDArray[np.float64], float],
    fractal_prefactor: Union[NDArray[np.float64], float],
) -> Union[NDArray[np.float64], float]:
    """Collision radius according to Zurita-Gotor and Rosner (2002).

    Args:
        radius_gyration: Radius of gyration of the particle [m].
        fractal_prefactor: Fractal prefactor of the particle
            [dimensionless, k0].

    Returns:
        (float or NDArray[float]): Collision radius of the particle [m].

    References:
        Zurita-Gotor, M., & Rosner, D. E. (2002). Effective diameters for
        collisions of fractal-like aggregates: Recommendations for improved
        aerosol coagulation frequency predictions. Journal of Colloid and
        Interface Science, 255(1).
        https://doi.org/10.1006/jcis.2002.8634

    Examples:
        ``` py title="Example"
        zurita_gotor_2002(1.5, 1.2)
        # 1.577
        ```
    """
    return 1.037 * (fractal_prefactor**0.077) * radius_gyration


def thajudeen_2012(
    fractal_dimension: float,
    number_of_particles: float,
    radius_gyration: Union[NDArray[np.float64], float],
    radius_monomer: float,
) -> Union[NDArray[np.float64], float]:
    """Collision radius according to Thajudeen et al. (2012).

    Args:
        fractal_dimension: Fractal dimension of the particle [dimensionless].
        number_of_particles: Number of particles in the aggregate.
        radius_gyration: Radius of gyration of the particle [m].
        radius_monomer: Monomer radius [m].

    Returns:
        (float): Collision radius of the particle [m].

    References:
        Thajudeen, T., Gopalakrishnan, R., & Hogan, C. J. (2012). The
        collision rate of nonspherical particles and aggregates for all
        diffusive knudsen numbers. Aerosol Science and Technology, 46(11).
        https://doi.org/10.1080/02786826.2012.701353

    Examples:
        ``` py title="Example"
        thajudeen_2012(2.5, 100, 1.5, 0.1)
        # 2.034
        ```
    """
    alpha1 = 0.253 * fractal_dimension**2 - 1.209 * fractal_dimension + 1.433
    alpha2 = -0.218 * fractal_dimension**2 + 0.964 * fractal_dimension - 0.180
    phi = 1 / (alpha1 * np.log(number_of_particles) + alpha2)
    radius_s_i = phi * radius_gyration
    radius_s_ii = (
        radius_monomer * (1.203 - 0.4315 / fractal_dimension) / 2
    ) * (4 * radius_s_i / radius_monomer) ** (
        0.8806 + 0.3497 / fractal_dimension
    )
    return radius_s_ii / 2


def qian_2022_rg(
    radius_gyration: Union[NDArray[np.float64], float],
    radius_monomer: float,
) -> Union[NDArray[np.float64], float]:
    """Fitted model using radius of gyration.

    Args:
        radius_gyration: radius of gyration [m].
        radius_monomer: monomer radius [m].

    Returns:
        (float): Collision radius of the particle [m].

    References:
        Qian, W., Kronenburg, A., Hui, X., Lin, Y., & Karsch, M. (2022).
        Effects of agglomerate characteristics on their collision kernels in
        the free molecular regime. Journal of Aerosol Science, 159.
        https://doi.org/10.1016/j.jaerosci.2021.105868

    Examples:
        ``` py title="Example"
        qian_2022_rg(1.5, 0.1)
        # 1.5036
        ```
    """
    coefficient = (0.973, 0.441)
    return (
        coefficient[0] * (radius_gyration / radius_monomer) + coefficient[1]
    ) * radius_monomer


def qian_2022_rg_df(
    fractal_dimension: Union[NDArray[np.float64], float],
    radius_gyration: Union[NDArray[np.float64], float],
    radius_monomer: float,
) -> Union[NDArray[np.float64], float]:
    """New model Rc with Rg and Df parameters.

    Args:
        fractal_dimension: Fractal dimension of the particle [-].
        radius_gyration: Scaled radius of gyration [m].
        radius_monomer: Monomer radius [m].

    Returns:
        (float): Collision radius of the particle [m].

    References:
        Qian, W., Kronenburg, A., Hui, X., Lin, Y., & Karsch, M. (2022).
        Effects of agglomerate characteristics on their collision kernels in
        the free molecular regime. Journal of Aerosol Science, 159.
        https://doi.org/10.1016/j.jaerosci.2021.105868

    Examples:
        ``` py title="Example"
        qian_2022_rg_df(2.5, 1.5, 0.1)
        # 1.661
        ```
    """
    coefficient = (0.882, 0.223, 0.387)
    return (
        coefficient[0]
        * (fractal_dimension ** coefficient[1])
        * (radius_gyration / radius_monomer)
        + coefficient[2]
    ) * radius_monomer


def qian_2022_rg_df_k0(
    fractal_dimension: float,
    fractal_prefactor: float,
    radius_gyration: Union[NDArray[np.float64], float],
    radius_monomer: float,
) -> Union[NDArray[np.float64], float]:
    """New model Rc with Rg, Df, and k0 parameters.

    Args:
        fractal_dimension: Fractal dimension of the particle [-].
        fractal_prefactor: Fractal prefactor of the particle [-].
        radius_gyration: radius of gyration [m].
        radius_monomer: monomer radius [m].

    Returns:
        (float): Collision radius of the particle [m].

    References:
        Qian, W., Kronenburg, A., Hui, X., Lin, Y., & Karsch, M. (2022).
        Effects of agglomerate characteristics on their collision kernels in
        the free molecular regime. Journal of Aerosol Science, 159.
        https://doi.org/10.1016/j.jaerosci.2021.105868

    Examples:
        ``` py title="Example"
        qian_2022_rg_df_k0(2.5, 1.2, 1.5, 0.1)
        # 1.832
        ```
    """
    coefficient = (0.777, 0.479, 0.000970, 0.267, -0.0790)
    return (
        coefficient[0]
        * (fractal_dimension ** coefficient[1])
        * (fractal_prefactor ** coefficient[2])
        * (radius_gyration / radius_monomer)
        + coefficient[3] * fractal_prefactor
        + coefficient[4]
    ) * radius_monomer


def qian_2022_rg_df_k0_a13(
    fractal_dimension: float,
    fractal_prefactor: float,
    shape_anisotropy: float,
    radius_gyration: Union[NDArray[np.float64], float],
    radius_monomer: float,
) -> Union[NDArray[np.float64], float]:
    """New model Rc with Rg, Df, k0, and A13 parameters.

    Args:
        fractal_dimension: Fractal dimension of the particle [-].
        fractal_prefactor: Fractal prefactor of the particle [-].
        shape_anisotropy: Parameter A13 for the model [-].
        radius_gyration: Radius of gyration [m].
        radius_monomer: Monomer radius [m].

    Returns:
        (float): Collision radius of the particle [m].

    References:
        Qian, W., Kronenburg, A., Hui, X., Lin, Y., & Karsch, M. (2022).
        Effects of agglomerate characteristics on their collision kernels in
        the free molecular regime. Journal of Aerosol Science, 159.
        https://doi.org/10.1016/j.jaerosci.2021.105868

    Examples:
        ``` py title="Example"
        qian_2022_rg_df_k0_a13(2.5, 1.2, 0.5, 1.5, 0.1)
        # 1.823
        ```
    """
    coefficient = (0.876, 0.363, -0.105, 0.421, -0.0360, -0.227)
    return (
        coefficient[0]
        * (fractal_dimension ** coefficient[1])
        * (fractal_prefactor ** coefficient[2])
        * (radius_gyration / radius_monomer)
        + coefficient[3] * fractal_prefactor
        + coefficient[4] * shape_anisotropy
        + coefficient[5]
    ) * radius_monomer



"""Atmosphere module for modeling gas mixtures in a given environment."""

from dataclasses import dataclass, field
from particula.gas.species import GasSpecies


@dataclass
class Atmosphere:
    """Represents a mixture of gas species under specific conditions.

    This class represents the atmospheric environment by detailing properties
    such as temperature and pressure, alongside a dynamic list of gas species
    present.

    Attributes:
        temperature: Temperature of the gas mixture in Kelvin.
        total_pressure: Total atmospheric pressure of the mixture inPascals.
        species: List of GasSpecies objects representing the
            various species within the gas mixture.

    Methods:
        add_species(self, species: GasSpecies) -> None:
            Adds a GasSpecies object to the mixture.
        remove_species(self, index: int) -> None:
            Removes a GasSpecies object from the mixture based on its index.
    """

    temperature: float
    total_pressure: float
    species: list[GasSpecies] = field(default_factory=list)

    def add_species(self, gas_species: GasSpecies) -> None:
        """Adds a GasSpecies object to the mixture.

        Args:
            gas_species: The gas species to be added.
        """
        self.species.append(gas_species)

    def remove_species(self, index: int) -> None:
        """Removes a gas species from the mixture by its index.

        Args:
            index: Index of the gas species to remove. Must be within
                        the current range of the list.

        Raises:
            IndexError: If the provided index is out of bounds.
        """
        if 0 <= index < len(self.species):
            self.species.pop(index)
        else:
            raise IndexError("No gas species at the provided index.")

    def __iter__(self):
        """
        Allows iteration over the species in the gas mixture.

        Returns:
            Iterator[GasSpecies]: An iterator over the gas species objects.
        """
        return iter(self.species)

    def __getitem__(self, index: int) -> GasSpecies:
        """Retrieves a gas species by index.

        Args:
            index: The index of the gas species to retrieve.

        Returns:
            GasSpecies: The gas species at the specified index.
        """
        return self.species[index]

    def __len__(self) -> int:
        """Returns the number of species in the gas mixture.

        Returns:
            int: The number of gas species in the mixture.
        """
        return len(self.species)

    def __str__(self) -> str:
        """Provides a string representation of the Atmosphere object.

        Returns:
            str: A string that includes the temperature, pressure, and a
                list of species in the mixture.
        """
        return (
            f"Gas mixture at {self.temperature} K and {self.total_pressure} Pa"
            f" consisting of {[str(species) for species in self.species]}"
        )



"""
Test the sedimentation kernel functions.
"""

import numpy as np
from particula.dynamics.coagulation.sedimentation_kernel import (
    sedimentation_sp2016,
    sedimentation_sp2016_via_system_state,
)


def test_sedimentation_kernel():
    """Test just the sedimentation kernel."""
    radius_particle = np.array([1e-6, 2e-6, 3e-6, 4e-6])
    settling_velocities = np.array([0.01, 0.02, 0.03, 0.04])
    expected_kernel = np.array(
        [
            [0.0, 1.25663706e-13, 2.51327412e-13, 3.76991118e-13],
            [1.25663706e-13, 0.0, 1.25663706e-13, 2.51327412e-13],
            [2.51327412e-13, 1.25663706e-13, 0.0, 1.25663706e-13],
            [3.76991118e-13, 2.51327412e-13, 1.25663706e-13, 0.0],
        ]
    )

    kernel = sedimentation_sp2016(
        radius_particle=radius_particle,
        settling_velocities=settling_velocities,
        calculate_collision_efficiency=False,
    )

    assert np.allclose(kernel, expected_kernel, rtol=1e-6)


def test_sedimentation_kernel_via_system_state():
    """Test the sedimentation kernel via system state."""
    radius_particle = np.array([1e-6, 2e-6, 3e-6, 4e-6])
    density_particle = np.array([1000, 1000, 1000, 1000])
    temperature = 298.15
    pressure = 101325
    expected_kernel = np.array(
        [
            [0.0, 1.25663706e-13, 2.51327412e-13, 3.76991118e-13],
            [1.25663706e-13, 0.0, 1.25663706e-13, 2.51327412e-13],
            [2.51327412e-13, 1.25663706e-13, 0.0, 1.25663706e-13],
            [3.76991118e-13, 2.51327412e-13, 1.25663706e-13, 0.0],
        ]
    )

    kernel = sedimentation_sp2016_via_system_state(
        radius_particle=radius_particle,
        density_particle=density_particle,
        temperature=temperature,
        pressure=pressure,
        calculate_collision_efficiency=False,
    )

    assert np.allclose(kernel, expected_kernel, rtol=1e-6)
